from __future__ import unicode_literals

import logging

from django.contrib.auth.models import User
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from jsonfield import JSONField

from .utils import get_full_name
from .signals import (
    user_confirmed,
    user_rejected,
    user_activated,
    user_disabled,
    user_enabled
)

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class Profile(models.Model):

    user = models.OneToOneField(User)
    is_pending = models.BooleanField(default=False)
    is_confirmed = models.BooleanField(default=False)
    details = JSONField(null=True, blank=True)
    attributes = JSONField(null=True, blank=True)

    class Meta:
        ordering = ('user__last_name', 'user__last_name', 'user__username')

        verbose_name = _('Profile')
        verbose_name_plural = _('Profiles')

        permissions = (('view_profile', 'Can view Profile'),)

    def __str__(self):
        return self.user.username

    @property
    def full_name(self):
        return get_full_name(self.user)

    def confirm(self, request):
        self.is_confirmed = True
        self.save()

        user_confirmed.send(sender=self.__class__, request=request, user=self.user)

    def reject(self, request):
        self.is_pending = False
        self.user.is_active = False
        self.save()

        user_rejected.send(sender=self.__class__, request=request, user=self.user)


    def activate(self, request):
        self.is_confirmed = True
        self.is_pending = False
        self.save()

        user_activated.send(sender=self.__class__, request=request, user=self.user)


    def disable(self, request):
        self.user.is_active = False
        self.user.save(update_fields=['is_active'])

        user_disabled.send(sender=self.__class__, request=request, user=self.user)


    def enable(self, request):
        self.user.is_active = True
        self.user.save(update_fields=['is_active'])

        user_enabled.send(sender=self.__class__, request=request, user=self.user)



from django.contrib import admin

from .models import Database, Table, Column, Function


class DatabaseAdmin(admin.ModelAdmin):
    search_fields = ('__str__', )
    list_display = ('order' ,'__str__', 'access_level', 'metadata_access_level')
    list_display_links = ('__str__', )


class TableAdmin(admin.ModelAdmin):
    search_fields = ('__str__', )
    list_display = ('order' ,'__str__', 'access_level', 'metadata_access_level')
    list_display_links = ('__str__', )


class ColumnAdmin(admin.ModelAdmin):
    search_fields = ('__str__', 'datatype')
    list_display = ('order' ,'__str__', 'datatype', 'access_level', 'metadata_access_level')
    list_display_links = ('__str__', )


class FunctionAdmin(admin.ModelAdmin):
    search_fields = ('name', 'query_string')
    list_display = ('order', 'name', 'query_string')
    list_display_links = ('name', )


admin.site.register(Database, DatabaseAdmin)
admin.site.register(Table, TableAdmin)
admin.site.register(Column, ColumnAdmin)
admin.site.register(Function, FunctionAdmin)



from kombu import Exchange, Queue

CELERY_BROKER_URL = 'amqp://'
CELERY_RESULT_BACKEND = 'redis://localhost:6379/0'

CELERY_TASK_ACKS_LATE = True
CELERY_WORKER_PREFETCH_MULTIPLIER = 1

CELERY_TASK_DEFAULT_QUEUE = 'default'

CELERY_TASK_QUEUES = [
    Queue('default', Exchange('default'), routing_key='default'),
    Queue('download', Exchange('download'), routing_key='download'),
    Queue('query', Exchange('query'), routing_key='query', queue_arguments={
        'x-max-priority': 5
    })
]



from daiquiri.core.utils import send_mail
from daiquiri.auth.utils import get_admin_emails


def send_contact_message(request, message):

    emails = get_admin_emails()
    context = {
        'user': message.user,
        'author': message.author,
        'email': message.email,
        'subject': message.subject,
        'message': message.message,
    }
    send_mail(request, 'contact/email/new_message_admin', context, emails)
    send_mail(request, 'contact/email/new_message_user', context, [message.email])



from rest_framework import filters

from .models import ContactMessage


class SpamBackend(filters.BaseFilterBackend):

    def filter_queryset(self, request, queryset, view):
        spam = request.GET.get('spam')

        if spam is not None:
            if spam == 'true':
                queryset = queryset.filter(status=ContactMessage.STATUS_SPAM)
            else:
                queryset = queryset.exclude(status=ContactMessage.STATUS_SPAM)

        return queryset



from collections import OrderedDict

from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.exceptions import ParseError, NotFound
from rest_framework.reverse import reverse

from daiquiri.core.adapter import get_adapter

from .serializers import ColumnSerializer
from .utils import get_columns


class RowViewSet(viewsets.ViewSet):

    def list(self, request, *args, **kwargs):

        # get database and table from the querystring
        database_name = self.request.GET.get('database')
        table_name = self.request.GET.get('table')
        column_name = self.request.GET.get('column')

        # get the ordering
        ordering = self.request.GET.get('ordering')

        # get the search string
        filter_string = self.request.GET.get('filter')

        # get the page from the querystring and make sure it is an int
        page = self._get_page()

        # get the page_size from the querystring and make sure it is an int
        page_size = self._get_page_size()

        # get the columns using the utils function
        columns = get_columns(self.request.user, database_name, table_name)

        if columns:
            column_names = [column['name'] for column in columns]

            if column_name:
                if column_name not in column_names:
                    raise NotFound()
                else:
                    column_names = [column_name]

            # get database adapter
            adapter = get_adapter()

            # query the database for the total number of rows
            count = adapter.database.count_rows(database_name, table_name, column_names, filter_string)

            # query the paginated rowset
            results = adapter.database.fetch_rows(database_name, table_name, column_names, ordering, page, page_size, filter_string)

            # flatten the list if only one column is retrieved
            if column_name:
                results = [element for row in results for element in row]

            # get the previous and next url
            next = self._get_next_url(page, page_size, count)
            previous = self._get_previous_url(page)

            # return ordered dict to be send as json
            return Response(OrderedDict((
                ('count', count),
                ('next', next),
                ('previous', previous),
                ('results', results)
            )))

        # if nothing worked, return 404
        raise NotFound()

    def _get_page(self):
        try:
            return int(self.request.GET.get('page', '1'))
        except ValueError:
            raise ParseError('page must be an integer')

    def _get_page_size(self):
        try:
            return int(self.request.GET.get('page_size', '10'))
        except ValueError:
            raise ParseError('page_size must be an integer')

    def _get_next_url(self, page, page_size, count):
        url = reverse('serve:row-list', request=self.request)
        querydict = self.request.GET.copy()

        if page * page_size < count:
            querydict['page'] = page + 1
            return url + '?' + querydict.urlencode()
        else:
            return None

    def _get_previous_url(self, page):
        url = reverse('serve:row-list', request=self.request)
        querydict = self.request.GET.copy()

        if page > 1:
            querydict['page'] = page - 1
            return url + '?' + querydict.urlencode()
        else:
            return None


class ColumnViewSet(viewsets.ViewSet):

    def list(self, request, *args, **kwargs):

        # get database and table from the querystring
        database_name = self.request.GET.get('database')
        table_name = self.request.GET.get('table')

        # get the columns using the utils function
        columns = get_columns(self.request.user, database_name, table_name)

        if columns:
            return Response(ColumnSerializer(columns, many=True).data)

        # if nothing worked, return 404
        raise NotFound()



import logging
import re
import warnings

from django.db import OperationalError, ProgrammingError

from .base import DatabaseAdapter

logger = logging.getLogger(__name__)


class MySQLAdapter(DatabaseAdapter):

    FUNCTIONS = (
        # group_functions
        'AVG', 'COUNT', 'MAX_SYM', 'MIN_SYM', 'SUM', 'BIT_AND', 'BIT_OR', 'BIT_XOR',
        'BIT_COUNT', 'GROUP_CONCAT', 'STD', 'STDDEV', 'STDDEV_POP', 'STDDEV_SAMP',
        'VAR_POP', 'VAR_SAMP', 'VARIANCE',
        # number_functions
        'ABS', 'ACOS', 'ASIN', 'ATAN2', 'ATAN', 'CEIL', 'CEILING', 'CONV', 'COS', 'COT',
        'CRC32', 'DEGREES', 'EXP', 'FLOOR', 'LN', 'LOG10', 'LOG2', 'LOG', 'MOD', 'PI', 'POW',
        'POWER', 'RADIANS', 'RAND', 'ROUND', 'SIGN', 'SIN', 'SQRT', 'TAN', 'TRUNCATE',
        # time_functions
        'ADDDATE', 'ADDTIME', 'CONVERT_TZ', 'CURDATE', 'CURTIME', 'DATE_ADD',
        'DATE_FORMAT', 'DATE_SUB', 'DATE_SYM', 'DATEDIFF', 'DAYNAME', 'DAYOFMONTH',
        'DAYOFWEEK', 'DAYOFYEAR', 'EXTRACT', 'FROM_DAYS', 'FROM_UNIXTIME', 'GET_FORMAT',
        'HOUR', 'LAST_DAY', 'MAKEDATE', 'MAKETIME', 'MICROSECOND', 'MINUTE', 'MONTH',
        'MONTHNAME', 'NOW', 'PERIOD_ADD', 'PERIOD_DIFF', 'QUARTER', 'SEC_TO_TIME',
        'SECOND', 'STR_TO_DATE', 'SUBTIME', 'SYSDATE', 'TIME_FORMAT', 'TIME_TO_SEC',
        'TIME_SYM', 'TIMEDIFF', 'TIMESTAMP', 'TIMESTAMPADD', 'TIMESTAMPDIFF', 'TO_DAYS',
        'TO_SECONDS', 'UNIX_TIMESTAMP', 'UTC_DATE', 'UTC_TIME', 'UTC_TIMESTAMP', 'WEEK',
        'WEEKDAY', 'WEEKOFYEAR', 'YEAR', 'YEARWEEK',
    )

    DATATYPES = {
        'char': {
            'datatype': 'char',
            'arraysize': True
        },
        'varchar': {
            'datatype': 'char',
            'arraysize': True
        },
        'tinyint': {
            'datatype': 'unsignedByte',
            'arraysize': False
        },
        'smallint': {
            'datatype': 'short',
            'arraysize': False
        },
        'int': {
            'datatype': 'int',
            'arraysize': False
        },
        'bigint': {
            'datatype': 'long',
            'arraysize': False
        },
        'float': {
            'datatype': 'float',
            'arraysize': False
        },
        'double': {
            'datatype': 'double',
            'arraysize': False
        },
        'timestamp': {
            'datatype': 'timestamp',
            'arraysize': False
        }
    }

    def fetch_pid(self):
        return self.connection().connection.thread_id()

    def escape_identifier(self, identifier):
        # escape backticks whithin the identifier and backtick the string
        return '`%s`' % identifier.replace('`', '``')

    def escape_string(self, string):
        return "'%s'" % string

    def build_query(self, database_name, table_name, query, timeout, max_records):
        # construct the actual query
        params = {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name),
            'query': query,
            'timeout': timeout,
            'max_records': max_records
        }

        if max_records is not None:
            return 'CREATE TABLE %(database)s.%(table)s ENGINE=MyISAM ( %(query)s ) LIMIT %(max_records)s;' % params
        else:
            return 'CREATE TABLE %(database)s.%(table)s ENGINE=MyISAM ( %(query)s );' % params

    def abort_query(self, pid):
        sql = 'KILL %(pid)i' % {'pid': pid}
        self.execute(sql)

    def count_rows(self, database_name, table_name, column_names=None, filter_string=None):
        # if no column names are provided get all column_names from the table
        if not column_names:
            column_names= self.fetch_column_names(database_name, table_name)

        # prepare sql string
        sql = 'SELECT COUNT(*) FROM %(database)s.%(table)s' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name)
        }

        # process filtering
        if filter_string:
            # create a list of escaped columns
            escaped_column_names = [self.escape_identifier(column_name) for column_name in column_names]

            sql_args = []
            where_stmts = []
            for escaped_column_name in escaped_column_names:
                sql_args.append('%' + filter_string + '%')
                where_stmts.append(escaped_column_name + ' LIKE %s')

            sql += ' WHERE ' + ' OR '.join(where_stmts)
        else:
            sql_args = None

        return self.fetchone(sql, args=sql_args)[0]

    def fetch_stats(self, database_name, table_name):
        sql = 'SELECT table_rows as nrows, data_length + index_length AS size FROM `information_schema`.`tables` WHERE `table_schema` = %s AND table_name = %s;'
        return self.fetchone(sql, (database_name, table_name))

    def fetch_rows(self, database_name, table_name, column_names=None, ordering=None, page=1, page_size=10, filter_string=None):
        # if no column names are provided get all column_names from the table
        if not column_names:
            column_names= self.fetch_column_names(database_name, table_name)

        # create a list of escaped columns
        escaped_column_names = [self.escape_identifier(column_name) for column_name in column_names]

        # prepare sql string
        sql = 'SELECT %(columns)s FROM %(database)s.%(table)s' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name),
            'columns': ', '.join(escaped_column_names)
        }

        # process filtering
        if filter_string:
            sql_args = []
            where_stmts = []
            for escaped_column_name in escaped_column_names:
                sql_args.append('%' + filter_string + '%')
                where_stmts.append(escaped_column_name + ' LIKE %s')

            sql += ' WHERE ' + ' OR '.join(where_stmts)
        else:
            sql_args = None

        # process ordering
        if ordering:
            if ordering.startswith('-'):
                escaped_ordering_column, ordering_direction = self.escape_identifier(ordering[1:]), 'DESC'
            else:
                escaped_ordering_column, ordering_direction = self.escape_identifier(ordering), 'ASC'

            if escaped_ordering_column in escaped_column_names:
                sql += ' ORDER BY %(column)s %(direction)s' % {
                    'column': escaped_ordering_column,
                    'direction': ordering_direction
                }

        # process page and page_size
        if page_size > 0:
            sql += ' LIMIT %(limit)s OFFSET %(offset)s' % {
                'limit': page_size,
                'offset': (int(page) - 1) * int(page_size)
            }

        return self.fetchall(sql, args=sql_args)

    def fetch_row(self, database_name, table_name, column_name, value):
        # prepare sql string
        sql = 'SELECT * FROM %(database)s.%(table)s WHERE %(column)s = %%s' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name),
            'column': self.escape_identifier(column_name)
        }

        return self.fetchone(sql, args=(value, ))

    def create_user_database_if_not_exists(self, database_name):
        # escape input
        escaped_database_name = self.escape_identifier(database_name)

        # prepare sql string
        sql = 'CREATE DATABASE IF NOT EXISTS %(database)s' % {
            'database': escaped_database_name
        }

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.execute(sql)


    def fetch_tables(self, database_name):
        # escape input
        escaped_database_name = self.escape_identifier(database_name)

        # prepare sql string
        sql = 'SHOW FULL TABLES FROM %(database)s' % {
            'database': escaped_database_name
        }

        # execute query
        try:
            rows = self.fetchall(sql)
        except OperationalError as e:
            logger.error('Could not fetch from %s (%s)' % (database_name, e))
            return []
        else:
            return [{
                'name': row[0],
                'type': 'view' if row[1] == 'VIEW' else 'table'
            } for row in rows]

    def fetch_table(self, database_name, table_name):
        # prepare sql string
        sql = 'SHOW FULL TABLES FROM %(database)s LIKE %(table)s' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_string(table_name)
        }

        # execute query
        try:
            row = self.fetchone(sql)
        except OperationalError as e:
            logger.error('Could not fetch %s.%s (%s)' % (database_name, table_name, e))
            return {}
        else:
            return {
                'name': row[0],
                'type': 'view' if row[1] == 'VIEW' else 'table'
            }

    def rename_table(self, database_name, table_name, new_table_name):
        sql = 'RENAME TABLE %(database)s.%(table)s to %(database)s.%(new_table)s;' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name),
            'new_table': self.escape_identifier(new_table_name)
        }

        self.execute(sql)

    def drop_table(self, database_name, table_name):
        sql = 'DROP TABLE IF EXISTS %(database)s.%(table)s;' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name)
        }

        self.execute(sql)

    def fetch_columns(self, database_name, table_name):
        # prepare sql string
        sql = 'SHOW FULL COLUMNS FROM %(database)s.%(table)s;' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name)
        }

        # execute query
        try:
            rows = self.fetchall(sql)
        except ProgrammingError as e:
            logger.error('Could not fetch from %s.%s (%s)' % (database_name, table_name, e))
            return []
        else:
            column_metadata = []
            for row in rows:
                datatype, arraysize = self.convert_datatype(row[1])

                column_metadata.append({
                    'name': row[0],
                    'datatype': datatype,
                    'arraysize': arraysize,
                    'indexed': bool(row[4])
                })

            return column_metadata

    def fetch_column(self, database_name, table_name, column_name):
        # prepare sql string
        sql = 'SHOW FULL COLUMNS FROM %(database)s.%(table)s WHERE `Field` = %(column)s' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name),
            'column': self.escape_string(column_name)
        }

        # execute query
        try:
            row = self.fetchone(sql)
        except ProgrammingError as e:
            logger.error('Could not fetch %s.%s.%s (%s)' % (database_name, table_name, column_name, e))
            return {}
        else:
            return {
                'name': row[0],
                'datatype': row[1],
                'indexed': bool(row[4])
            }

    def fetch_column_names(self, database_name, table_name):
        # prepare sql string
        sql = 'SHOW COLUMNS FROM %(database)s.%(table)s' % {
            'database': self.escape_identifier(database_name),
            'table': self.escape_identifier(table_name),
        }
        return [column[0] for column in self.fetchall(sql)]

    def convert_datatype(self, datatype_string):
        result = re.match('([a-z]+)\(*(\d*)\)*', datatype_string)

        if result:
            native_datatype = result.group(1)

            try:
                native_arraysize = int(result.group(2))
            except ValueError:
                native_arraysize = None

            if native_datatype in self.DATATYPES:
                datatype = self.DATATYPES[native_datatype]['datatype']

                if self.DATATYPES[native_datatype]['arraysize']:
                    arraysize = native_arraysize
                else:
                    arraysize = None

                return datatype, arraysize
            else:
                return native_datatype, native_arraysize
        else:
            return datatype_string, None



from django.test import TestCase

from test_generator.views import TestViewMixin


class FilesViewTestCase(TestCase):

    fixtures = (
        'auth.json',
        'files.json'
    )

    users = (
        ('admin', 'admin'),
        ('manager', 'manager'),
        ('user', 'user'),
        ('anonymous', None),
    )


class InternalFileTests(TestViewMixin, FilesViewTestCase):

    url_names = {
        'list_view': 'files:file'
    }

    status_map = {
        'list_view': {
            'admin': 200, 'manager': 200, 'user': 200, 'anonymous': 404
        }
    }

    def _test_get(self, username):
        self.assert_list_view(username, kwargs={
            'file_path': 'html/'
        })


class PrivateFileTests(TestViewMixin, FilesViewTestCase):

    url_names = {
        'list_view': 'files:file'
    }

    status_map = {
        'list_view': {
            'admin': 404, 'manager': 200, 'user': 404, 'anonymous': 404
        }
    }

    def _test_get(self, username):
        self.assert_list_view(username, kwargs={
            'file_path': 'image_00.jpg'
        })



from sendfile import sendfile

from rest_framework import viewsets
from rest_framework.exceptions import NotFound

from .utils import search_file


class FileViewSet(viewsets.GenericViewSet):

    def list(self, request):

        search = request.GET.get('search', None)

        if search:
            file_name = search_file(request.user, search)

            if file_name:
                return sendfile(request, file_name, attachment=False)
            else:
                raise NotFound()

        else:
            raise NotFound()



import os
import json
import logging
import six

from collections import OrderedDict

from celery.result import AsyncResult, EagerResult
from celery.task.control import revoke

from django.conf import settings
from django.contrib.auth.models import Group
from django.db import models
from django.db.utils import OperationalError, ProgrammingError
from django.urls import reverse
from django.utils.encoding import python_2_unicode_compatible
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from rest_framework.exceptions import ValidationError

from jsonfield import JSONField

from queryparser.mysql import MySQLQueryProcessor
from queryparser.adql import ADQLQueryTranslator
from queryparser.exceptions import QueryError, QuerySyntaxError

from daiquiri.core.adapter import get_adapter
from daiquiri.core.constants import ACCESS_LEVEL_CHOICES
from daiquiri.jobs.models import Job

from .managers import QueryJobManager, ExampleManager
from .utils import (
    get_quota,
    get_default_table_name,
    get_default_queue,
    get_tap_schema_name,
    get_user_database_name,
    get_download_file_name,
    get_asterisk_columns,
    check_permissions
)
from .tasks import run_query, create_download_file

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class QueryJob(Job):

    objects = QueryJobManager()

    database_name = models.CharField(max_length=256)
    table_name = models.CharField(max_length=256)
    queue = models.CharField(max_length=16)

    query_language = models.CharField(max_length=16)
    query = models.TextField()
    native_query = models.TextField(null=True, blank=True)
    actual_query = models.TextField(null=True, blank=True)

    queue = models.CharField(max_length=16, null=True, blank=True)
    nrows = models.IntegerField(null=True, blank=True)
    size = models.IntegerField(null=True, blank=True)

    metadata = JSONField()

    pid = models.IntegerField(null=True, blank=True)

    class Meta:
        ordering = ('start_time', )

        verbose_name = _('QueryJob')
        verbose_name_plural = _('QueryJobs')

        permissions = (('view_queryjob', 'Can view QueryJob'),)

    def __str__(self):
        return self.get_str()

    def save(self, *args, **kwargs):
        if self.pk is None:
            self.phase = self.PHASE_PENDING
            self.creation_time = now()

        return super(QueryJob, self).save(*args, **kwargs)

    def process(self):
        if not self.response_format:
            self.response_format = settings.QUERY_DEFAULT_DOWNLOAD_FORMAT

        # set the database name
        if not self.database_name:
            self.database_name = get_user_database_name(self.owner)

        # set a default table name
        if not self.table_name:
            self.table_name = get_default_table_name()

        if not self.query_language:
            raise ValidationError({
                 'query_language': [_('This field may not be blank.')]
            })

        if not self.query:
            raise ValidationError({
                 'query': [_('This field may not be blank.')]
            })

        # check quota
        if QueryJob.objects.get_size(self.owner) > get_quota(self.owner):
            raise ValidationError({
                    'query': [_('Quota is exceeded. Please remove some of your jobs.')]
                })

        # translate adql -> mysql string
        if self.query_language == 'adql-2.0':
            try:
                translator = ADQLQueryTranslator(self.query)
                translated_query = translator.to_mysql()
            except QuerySyntaxError as e:
                raise ValidationError({
                    'query': {
                        'messages': [_('There has been an error while parsing your query.')],
                        'positions': json.dumps(e.syntax_errors),
                    }
                })
            except QueryError as e:
                raise ValidationError({
                    'query': {
                        'messages': e.messages,
                    }
                })

        else:
            translated_query = self.query

        # parse the query
        try:
            processor = MySQLQueryProcessor(translated_query)

            tap_schema_name = get_tap_schema_name()
            if tap_schema_name:
                processor.process_query(replace_schema_name={
                    'TAP_SCHEMA': tap_schema_name
                })
            else:
                processor.process_query()

        except QuerySyntaxError as e:
            raise ValidationError({
                'query': {
                    'messages': [_('There has been an error while parsing your query.')],
                    'positions': json.dumps(e.syntax_errors),
                }
            })
        except QueryError as e:
            raise ValidationError({
                'query': {
                    'messages': e.messages,
                }
            })

        # check permissions
        permission_messages = check_permissions(self.owner, processor.keywords, processor.columns, processor.functions)
        if permission_messages:
            raise ValidationError({
                'query': permission_messages
            })

        # process display_columns to expand *
        display_columns = []
        for display_column in processor.display_columns:
            if display_column[0] == '*':
                display_columns += get_asterisk_columns(display_column)
            else:
                display_columns.append(display_column)

        # check for duplicate columns in display_columns
        display_column_names = [column_name for column_name, column in display_columns]
        seen = set()
        duplicate_columns = []
        for column_name in display_column_names:
            if column_name not in seen:
                seen.add(column_name)
            else:
                duplicate_columns.append(column_name)

        if duplicate_columns:
            raise ValidationError({
                'query': [_('Duplicate column name \'%(column)s\'') % {
                    'column': duplicate_column
                } for duplicate_column in duplicate_columns]
            })

        # initialize metadata and store map of aliases
        self.metadata = {
            'display_columns': OrderedDict(display_columns)
        }

        # get the native query from the processor (without trailing semicolon)
        self.native_query = processor.query.rstrip(';')

        # set clean flag
        self.is_clean = True


    @property
    def parameters(self):
        return {
            'database_name': self.database_name,
            'table_name': self.table_name,
            'query_language': self.query_language,
            'query': self.query,
            'native_query': self.native_query,
            'actual_query': self.actual_query,
            'queue': self.queue,
            'nrows': self.nrows,
            'size': self.size
        }

    @property
    def timeout(self):
        if self.queue:
            return six.next((queue['timeout'] for queue in settings.QUERY_QUEUES if queue['key'] == self.queue))
        else:
            return 10

    @property
    def priority(self):
        return six.next((queue['priority'] for queue in settings.QUERY_QUEUES if queue['key'] == self.queue))

    @property
    def result_status(self):
        return 'OK' if self.max_records is None else 'OVERFLOW'

    @property
    def results(self):
        if self.phase == self.PHASE_COMPLETED:
            # create dictionary of the form
            # 'votable': '/query/api/jobs/{job.id}/stream/votable'
            return {download_format['key']: reverse('query:job-stream', kwargs={
                'pk': str(self.id),
                'format_key': download_format['key']
            }) for download_format in settings.QUERY_DOWNLOAD_FORMATS}
        else:
            return {}

    @property
    def result(self):
        return reverse('query:job-stream', kwargs={
            'pk': str(self.id),
            'format_key': self.response_format
        })

    @property
    def quote(self):
        return None

    @property
    def time_queue(self):
        if self.start_time and self.creation_time:
            return (self.start_time - self.creation_time).total_seconds()
        else:
            return None

    @property
    def time_query(self):
        if self.end_time and self.start_time:
            return (self.end_time - self.start_time).total_seconds()
        else:
            return None

    def run(self, sync=False):
        if not self.is_clean:
            raise Exception('job.process() was not called.')

        if self.phase == self.PHASE_PENDING:
            self.phase = self.PHASE_QUEUED
            self.save()

            # start the submit_query task in a syncronous or asuncronous way
            job_id = str(self.id)
            if not settings.ASYNC or sync:
                logger.info('run_query %s submitted (sync)' % self.id)
                run_query.apply((job_id, ), task_id=job_id, throw=True)

            else:
                if not self.queue:
                    self.queue = get_default_queue()
                    self.save()

                logger.info('run_query %s submitted (async, queue=query, priority=%s)' % (self.id, self.priority))
                run_query.apply_async((job_id, ), task_id=job_id, queue='query', priority=self.priority)

        else:
            raise ValidationError({
                'phase': ['Job is not PENDING.']
            })

    def abort(self):
        if settings.ASYNC:
            # first, revoke the task in celery, regardless the phase
            revoke(str(self.id))

        current_phase = self.phase

        if current_phase in self.PHASE_ACTIVE:
            # next, set the phase to ABORTED
            self.phase = self.PHASE_ABORTED
            self.save()

            # finally, abort query, this will trigger OperationalError in the run_query task
            if current_phase == self.PHASE_EXECUTING:
                self.abort_query()

    def archive(self):
        self.abort()
        self.drop_table()
        self.phase = self.PHASE_ARCHIVED
        self.nrows = None
        self.size = None
        self.save()

    def rename_table(self, new_table_name):
        if self.table_name != new_table_name:
            get_adapter().database.rename_table(self.database_name, self.table_name, new_table_name)

            self.metadata['name'] = new_table_name
            self.save()

    def drop_table(self):
        # drop the corresponding database table, but fail silently
        try:
            get_adapter().database.drop_table(self.database_name, self.table_name)
        except ProgrammingError:
            pass

    def abort_query(self):
        # abort the job on the database
        try:
            get_adapter().database.abort_query(self.pid)
        except OperationalError:
            # the query was probably killed before
            pass

    def download(self, format):
        if self.phase == self.PHASE_COMPLETED:
            file_name = get_download_file_name(self.owner, self.table_name, format)

            task_id = file_name
            task_args = (file_name, format['key'], self.database_name, self.table_name, self.metadata, self.result_status, (self.nrows == 0))

            # create directory if necessary
            try:
                os.mkdir(os.path.dirname(file_name))
            except OSError:
                pass

            if not settings.ASYNC:
                if os.path.isfile(file_name):
                    task_result = EagerResult(task_id, None, 'SUCCESS')
                else:
                    logger.info('create_download_file %s submitted (sync)' % self.id)
                    task_result = create_download_file.apply(task_args, task_id=task_id, throw=True)
            else:
                task_result = AsyncResult(task_id)

                if not os.path.isfile(file_name):
                    # create an empty file to prevent multiple pending tasks
                    open(file_name, 'a').close()

                    if task_result.successful():
                        # somebody or something removed the file. start all over again
                        task_result.forget()

                    logger.info('create_download_file %s submitted (async, queue=download)' % self.id)
                    task_result = create_download_file.apply_async(task_args, task_id=task_id, queue='download')

            return task_result, file_name

        else:
            raise ValidationError({
                'phase': ['Job is not COMPLETED.']
            })

    def stream(self, format):
        if self.phase == self.PHASE_COMPLETED:
            return get_adapter().download.generate(format['key'], self.database_name, self.table_name, self.metadata, self.result_status, (self.nrows == 0))

        else:
            raise ValidationError({
                'phase': ['Job is not COMPLETED.']
            })

@python_2_unicode_compatible
class Example(models.Model):

    objects = ExampleManager()

    order = models.IntegerField(
        null=True, blank=True,
        verbose_name=_('Order'),
        help_text=_('Position in lists.')
    )
    name = models.CharField(
        max_length=256,
        verbose_name=_('Name'),
        help_text=_('Identifier of the example.')
    )
    description = models.TextField(
        null=True, blank=True,
        verbose_name=_('Description'),
        help_text=_('A brief description of the example to be displayed in the user interface.')
    )
    query_language = models.CharField(
        max_length=16,
        verbose_name=_('Query language'),
        help_text=_('The query language for this example.')
    )
    query_string = models.TextField(
        verbose_name=_('Query string'),
        help_text=_('The query string (SQL) for this example.')
    )
    access_level = models.CharField(
        max_length=8, choices=ACCESS_LEVEL_CHOICES,
        verbose_name=_('Access level')
    )
    groups = models.ManyToManyField(
        Group, blank=True,
        verbose_name=_('Groups'),
        help_text=_('The groups which have access to the examples.')
    )

    class Meta:
        ordering = ('order',)

        verbose_name = _('Example')
        verbose_name_plural = _('Examples')

        permissions = (('view_example', 'Can view Example'),)

    def __str__(self):
        return self.name



from django.core.urlresolvers import reverse
from django.test import TestCase

from ..models import Profile


class AccountsTestCase(TestCase):

    fixtures = (
        'auth.json',
    )

    def test_login(self):
        url = reverse('account_login')
        response = self.client.post(url, {
            'login': 'user',
            'password': 'user'
        })
        self.assertRedirects(response, reverse('home'))

    def test_invalid(self):
        url = reverse('account_login')
        response = self.client.post(url, {
            'login': 'invalid',
            'password': 'invalid'
        })
        self.assertEqual(response.status_code, 200)

    def test_logout(self):
        url = reverse('account_logout')
        response = self.client.post(url)
        self.assertRedirects(response, reverse('home'))

    def test_signup_get(self):
        url = reverse('account_signup')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_signup_post(self):
        url = reverse('account_signup')
        response = self.client.post(url, {
            'email': 'test@example.com',
            'username': 'test',
            'first_name': 'Tanja',
            'last_name': 'Test',
            'password1': 'testing',
            'password2': 'testing'
        })

        # check that the signup redirects to the pending page or the confirm email page
        self.assertEqual(response.status_code, 302)

        # check that a profile was created
        profile = Profile.objects.get(user__username='test')
        self.assertEqual(profile.is_pending, True)

    def test_signup_post_invalid(self):
        url = reverse('account_signup')
        response = self.client.post(url, {
            'email': 'test@example.com',
            'username': 'test',
            'first_name': 'Tanja',
            'last_name': 'Test',
            'password1': 'testing',
            'password2': 'invalid'
        })

        # check that the signup returns 200 (with validation error)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '* You must type the same password each time.')

        # check that a profile was not created
        exists = Profile.objects.filter(user__username='test').exists()
        self.assertEqual(exists, False)

    def test_signup_post_exists(self):
        url = reverse('account_signup')
        response = self.client.post(url, {
            'email': 'user@example.com',
            'username': 'user',
            'first_name': 'Tanja',
            'last_name': 'Test',
            'password1': 'testing',
            'password2': 'testing'
        })

        # check that the signup returns 200 (with validation error)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '* A user with that username already exists.')
        self.assertContains(response, '* A user is already registered with this e-mail address.')

        # check that a profile was not created
        exists = Profile.objects.filter(user__username='test').exists()
        self.assertEqual(exists, False)

    def test_profile_get_for_user(self):
        self.client.login(username='user', password='user')

        url = reverse('account_profile')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_profile_get_for_anonymous(self):
        url = reverse('account_profile')
        response = self.client.get(url)
        self.assertRedirects(response, reverse('account_login') + '?next=' + url)

    def test_profile_post_for_user(self):
        self.client.login(username='user', password='user')

        url = reverse('account_profile')
        response = self.client.post(url, {
            'first_name': 'Tanja',
            'last_name': 'Test',
        })
        self.assertRedirects(response, reverse('home'))

    def test_profile_cancel_for_user(self):
        self.client.login(username='user', password='user')

        url = reverse('account_profile')
        response = self.client.post(url, {
            'cancel': True
        })
        self.assertRedirects(response, reverse('home'))

    def test_profile_post_for_anonymous(self):
        url = reverse('account_profile')
        response = self.client.post(url, {
            'first_name': 'Tanja',
            'last_name': 'Test',
        })
        self.assertRedirects(response, reverse('account_login') + '?next=' + url)

    def test_profile_json_get_for_user(self):
        self.client.login(username='user', password='user')

        url = reverse('account_profile_json')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_profile_json_get_for_anonymous(self):
        url = reverse('account_profile_json')
        response = self.client.get(url)
        self.assertRedirects(response, reverse('account_login') + '?next=' + url)

    def test_token_get_for_user(self):
        self.client.login(username='user', password='user')

        url = reverse('account_token')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_token_get_for_anonymous(self):
        url = reverse('account_token')
        response = self.client.get(url)
        self.assertRedirects(response, reverse('account_login') + '?next=' + url)

    def test_token_post_for_user(self):
        self.client.login(username='user', password='user')

        url = reverse('account_token')
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)

    def test_token_post_for_anonymous(self):
        url = reverse('account_token')
        response = self.client.post(url)
        self.assertRedirects(response, reverse('account_login') + '?next=' + url)



from django.conf import settings
from django.contrib.auth.mixins import AccessMixin
from django.views.generic import TemplateView

from daiquiri.core.views import ModelPermissionMixin
from daiquiri.core.utils import get_model_field_meta

from .models import Example


class QueryView(AccessMixin, TemplateView):
    template_name = 'query/query.html'

    def dispatch(self, request, *args, **kwargs):
        if not settings.QUERY_ANONYMOUS and not request.user.is_authenticated:
            return self.handle_no_permission()
        return super(QueryView, self).dispatch(request, *args, **kwargs)


class ExamplesView(ModelPermissionMixin, TemplateView):

    template_name = 'query/examples.html'
    permission_required = 'daiquiri_query.view_example'

    def get_context_data(self, **kwargs):
        context = super(ExamplesView, self).get_context_data(**kwargs)
        context['meta'] = {
            'Example': get_model_field_meta(Example)
        }
        return context



from django.contrib.auth.mixins import PermissionRequiredMixin as DjangoPermissionRequiredMixin
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import PermissionDenied
from django.shortcuts import render

from allauth.account.forms import LoginForm

from rules.contrib.views import PermissionRequiredMixin as RulesPermissionRequiredMixin


def home(request):
    if not request.user.is_authenticated():
        login_form = LoginForm()
    else:
        login_form = None

    return render(request, 'core/home.html', {'form': login_form})


class PermissionRedirectMixin(object):

    def handle_no_permission(self):
        if self.request.user.is_authenticated():
            raise PermissionDenied(self.get_permission_denied_message())

        return redirect_to_login(self.request.get_full_path(), self.get_login_url(), self.get_redirect_field_name())


class ModelPermissionMixin(PermissionRedirectMixin, DjangoPermissionRequiredMixin, object):
    pass


class ObjectPermissionMixin(PermissionRedirectMixin, RulesPermissionRequiredMixin, object):
    pass



from daiquiri.jobs.renderers import XMLRenderer


class ExampleRenderer(XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):

        self.start('div', {
            'vocab': 'http://www.ivoa.net/rdf/examples',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance'
        })
        for example in data:
            self.start('div', {
                'id': 'example-%i' % example['id'],
                'resource': '#example-%i' % example['id'],
                'typeof': 'example'
            })
            self.node('h2', {'property': 'name'}, example['name'])
            self.node('p', {'property': 'capability'}, 'ivo://ivoa.net/std/TAP')
            self.node('p', {'property': 'description'}, example['description'])
            self.node('pre', {'property': 'query'}, example['query_string'])
            self.end('div')

        self.end('div')

class AvailabilityRenderer(XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):

        self.start('vosi:availability', {
            'xmlns:vosi': 'http://www.ivoa.net/xml/VOSITables/v1.0',
            'xmlns:vs': 'http://www.ivoa.net/xml/VODataService/v1.0',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': 'http://www.ivoa.net/xml/VOSI/v1.0 http://www.ivoa.net/xml/VOSI/v1.0 http://www.ivoa.net/xml/VODataService/v1.0 http://www.ivoa.net/xml/VODataService/v1.0'
        })
        self.node('vosi:available', {}, data['available'])
        self.node('vosi:note', {}, data['note'])
        self.end('vosi:availability')


class CapabilitiesRenderer(XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):

        self.start('vosi:capabilities', {
            'xmlns:vosi': 'http://www.ivoa.net/xml/VOSITables/v1.0',
            'xmlns:vs': 'http://www.ivoa.net/xml/VODataService/v1.0',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': 'http://www.ivoa.net/xml/VOSI/v1.0 http://www.ivoa.net/xml/VOSI/v1.0 http://www.ivoa.net/xml/VODataService/v1.0 http://www.ivoa.net/xml/VODataService/v1.0'
        })

        for capability in data:

            self.start('vosi:capability', {'standardID': capability['schemaID']})

            self.start('interface', capability['interface']['attrs'])
            self.node('accessURL', capability['interface']['accessURL']['attrs'], capability['interface']['accessURL']['text'])
            self.end('interface')

            if 'languages' in capability:
                for language in capability['languages']:
                    self.start('language')
                    self.node('name', {}, language['name'])
                    self.node('version', {}, language['version'])
                    self.node('description', {}, language['description'])
                    self.end('language')

            self.end('vosi:capability')

        self.end('vosi:capabilities')


class TablesetRenderer(XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):
        self.start('vosi:tableset', {
            'xmlns:vosi': 'http://www.ivoa.net/xml/VOSITables/v1.0',
            'xmlns:vod': 'http://www.ivoa.net/xml/VODataService/v1.1',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': 'http://www.ivoa.net/xml/VODataService/v1.1 http://www.ivoa.net/xml/VOSITables/v1.0'
        })

        for schema in data:
            self.start('vosi:schema')
            self.node('name', {}, schema['name'])
            self.node('description', {}, schema['description'])

            for table in schema['tables']:
                self.start('vosi:table')
                self.node('name', {}, table['name'])
                self.node('description', {}, table['description'])

                for column in table['columns']:
                    self.start('vosi:column')
                    self.node('name', {}, column['name'])
                    self.node('dataType', {'xsi:type': 'vod:TAPType'}, column['datatype'])
                    self.node('ucd', {}, column['ucd'])
                    self.node('unit', {}, column['unit'])
                    for key in ['indexed', 'principal', 'std']:
                        if key in column and column[key]:
                            self.node('flag', {}, key)

                    self.end('vosi:column')

                self.end('vosi:table')

            self.end('vosi:schema')

        self.end('vosi:tableset')



from __future__ import unicode_literals

import uuid

from django.db import models
from django.contrib.auth.models import User
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _


@python_2_unicode_compatible
class Job(models.Model):

    PHASE_PENDING = 'PENDING'
    PHASE_QUEUED = 'QUEUED'
    PHASE_EXECUTING = 'EXECUTING'
    PHASE_COMPLETED = 'COMPLETED'
    PHASE_ERROR = 'ERROR'
    PHASE_ABORTED = 'ABORTED'
    PHASE_UNKNOWN = 'UNKNOWN'
    PHASE_HELD = 'HELD'
    PHASE_SUSPENDED = 'SUSPENDED'
    PHASE_ARCHIVED = 'ARCHIVED'
    PHASE_RUN = 'RUN'
    PHASE_ABORT = 'ABORT'
    PHASE_ACTIVE = (
        PHASE_PENDING,
        PHASE_QUEUED,
        PHASE_EXECUTING
    )
    PHASE_CHOICES = (
        (PHASE_PENDING, _('Pending')),
        (PHASE_QUEUED, _('Queued')),
        (PHASE_EXECUTING, _('Executing')),
        (PHASE_COMPLETED, _('Completed')),
        (PHASE_ERROR, _('Error')),
        (PHASE_ABORTED, _('Aborted')),
        (PHASE_UNKNOWN, _('Unknown')),
        (PHASE_HELD, _('Held')),
        (PHASE_SUSPENDED, _('Suspended')),
        (PHASE_ARCHIVED, _('Archived'))
    )

    JOB_TYPE_SYNC = 'SYNC'
    JOB_TYPE_ASYNC = 'ASYNC'
    JOB_TYPE_INTERFACE = 'INTERFACE'
    JOB_TYPE_CHOICES = (
        (JOB_TYPE_SYNC, _('Syncronous')),
        (JOB_TYPE_ASYNC, _('Asyncronous')),
        (JOB_TYPE_INTERFACE, _('Interface')),
    )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    owner = models.ForeignKey(User, blank=True, null=True)

    client_ip = models.GenericIPAddressField(blank=True, null=True)

    response_format = models.CharField(max_length=64, blank=True, null=True)
    max_records = models.IntegerField(blank=True, null=True)
    run_id = models.CharField(max_length=64, blank=True, null=True)

    phase = models.CharField(max_length=10, choices=PHASE_CHOICES)

    creation_time = models.DateTimeField(blank=True, null=True)
    start_time = models.DateTimeField(blank=True, null=True)
    end_time = models.DateTimeField(blank=True, null=True)
    execution_duration = models.PositiveIntegerField(blank=True, default=0)
    destruction_time = models.DateTimeField(blank=True, null=True)

    error_summary = models.CharField(max_length=256, blank=True, null=True)

    job_type = models.CharField(max_length=10, choices=JOB_TYPE_CHOICES)

    class Meta:
        ordering = ('start_time', )

        verbose_name = _('Job')
        verbose_name_plural = _('Jobs')

        permissions = (('view_job', 'Can view Job'),)

    def __str__(self):
        return self.get_str()

    def get_str(self):
        return "id=%s; phase=%s; job_type=%s" % (str(self.id), self.phase, self.job_type)

    @property
    def owner_username(self):
        return self.owner.username if self.owner else 'anonymous'

    @property
    def parameters(self):
        raise NotImplementedError

    @property
    def results(self):
        raise NotImplementedError

    @property
    def result(self):
        raise NotImplementedError

    @property
    def quote(self):
        raise NotImplementedError

    def run(self):
        raise NotImplementedError

    def abort(self):
        raise NotImplementedError

    def archive(self):
        raise NotImplementedError



from django.conf import settings

from rest_framework.permissions import BasePermission
from rest_framework.compat import is_authenticated


class HasPermission(BasePermission):

    def has_permission(self, request, view):
        return settings.QUERY_ANONYMOUS or (request.user and is_authenticated(request.user))



from django.test import TestCase
from django.contrib.auth.models import User, Group, Permission

from test_generator.viewsets import TestListViewsetMixin, TestDetailViewsetMixin, TestUpdateViewsetMixin

from ..models import Profile


class AuthViewsetTestCase(TestCase):

    fixtures = (
        'auth.json',
    )

    def setUp(self):
        group = Group.objects.get(name='managers')
        for codename in (
            'add_profile',
            'change_profile',
            'delete_profile',
            'view_profile'
        ):
            group.permissions.add(Permission.objects.get(codename=codename))


class ProfileTests(TestListViewsetMixin, TestDetailViewsetMixin, TestUpdateViewsetMixin, AuthViewsetTestCase):

    users = (
        ('admin', 'admin'),
        ('manager', 'manager'),
        ('user', 'user'),
        ('anonymous', None),
    )

    status_map = {
        'list_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'detail_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'create_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'update_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'delete_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'confirm_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'reject_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'activate_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'disable_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
        'enable_viewset': {
            'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 403
        },
    }

    instances = Profile.objects.all()
    url_names = {
        'viewset': 'auth:profile'
    }


    def _test_update_viewset(self, username):
        for instance in self.instances:
            instance.details = {}

            data = self.get_instance_as_dict(instance)
            data['user'] = {
                'first_name': instance.user.first_name,
                'last_name': instance.user.last_name,
                'groups': [group.id for group in instance.user.groups.all()]
            }

            self.assert_update_viewset(username, kwargs={
                'pk': instance.pk
            }, data=data)

    def _test_confirm_viewset(self, username):
        instance = User.objects.get(username='user')
        instance.profile.is_pending = True
        instance.save()

        self.assert_viewset('confirm_viewset', 'put', 'confirm', username, kwargs={
            'pk': instance.pk
        })

    def _test_reject_viewset(self, username):
        instance = User.objects.get(username='user')
        instance.profile.is_pending = True
        instance.save()

        self.assert_viewset('reject_viewset', 'put', 'reject', username, kwargs={
            'pk': instance.pk
        })

    def _test_activate_viewset(self, username):
        instance = User.objects.get(username='user')
        instance.profile.is_pending = True
        instance.profile.is_confirmed = True
        instance.save()

        self.assert_viewset('activate_viewset', 'put', 'activate', username, kwargs={
            'pk': instance.pk
        })

    def _test_disable_viewset(self, username):
        instance = User.objects.get(username='user')

        self.assert_viewset('disable_viewset', 'put', 'disable', username, kwargs={
            'pk': instance.pk
        })

    def _test_enable_viewset(self, username):
        instance = User.objects.get(username='user')
        instance.is_active = False
        instance.save()

        self.assert_viewset('enable_viewset', 'put', 'enable', username, kwargs={
            'pk': instance.pk
        })


class GroupViewSet(TestListViewsetMixin, TestDetailViewsetMixin, AuthViewsetTestCase):

    users = (
        ('admin', 'admin'),
        ('user', 'user'),
        ('anonymous', None)
    )

    instances = Group.objects.all()
    url_names = {
        'viewset': 'auth:group'
    }

    status_map = {
        'list_viewset': {
            'admin': 200, 'user': 403, 'anonymous': 403
        },
        'detail_viewset': {
            'admin': 200, 'user': 403, 'anonymous': 403
        }
    }



from rest_framework.reverse import reverse

def make_query_dict_upper_case(input_dict):
    output_dict = input_dict.copy()

    for key in output_dict.keys():
        if key.upper() != key:
            values = output_dict.getlist(key)

            if key.upper() in output_dict:
                output_dict.appendlist(key.upper(), values)
            else:
                output_dict.setlist(key.upper(), values)

            output_dict.pop(key)

    return output_dict

def get_job_url(request, kwargs):
    namespace = request.resolver_match.namespace
    base_name = request.resolver_match.url_name.rsplit('-', 1)[0]
    return reverse('%s:%s-detail' % (namespace, base_name), request=request, kwargs=kwargs)



from django.dispatch import receiver
from django.db.models.signals import post_save, post_delete

from daiquiri.metadata.models import Database, Table, Column

from .utils import (
    update_database,
    delete_database,
    update_table,
    delete_table,
    update_column,
    delete_column
)

@receiver(post_save, sender=Database)
def database_updated_handler(sender, **kwargs):
    update_database(kwargs['instance'])


@receiver(post_delete, sender=Database)
def database_deleted_handler(sender, **kwargs):
    delete_database(kwargs['instance'])


@receiver(post_save, sender=Table)
def table_updated_handler(sender, **kwargs):
    update_table(kwargs['instance'])


@receiver(post_delete, sender=Table)
def table_deleted_handler(sender, **kwargs):
    delete_table(kwargs['instance'])


@receiver(post_save, sender=Column)
def column_updated_handler(sender, **kwargs):
    update_column(kwargs['instance'])


@receiver(post_delete, sender=Column)
def column_deleted_handler(sender, **kwargs):
    delete_column(kwargs['instance'])



import logging

from sendfile import sendfile

from django.http import Http404

from .utils import get_file

logger = logging.getLogger(__name__)


def file(request, file_path):

    file_name = get_file(request.user, file_path)

    if file_name:
        return sendfile(request, file_name, attachment=False)

    # if nothing worked, return 404
    raise Http404



from django.contrib.auth.models import AnonymousUser
from django.test import TestCase, RequestFactory
from django.template import RequestContext, Template


class MetadataTagsTestCase(TestCase):

    fixtures = (
        'auth.json',
        'metadata.json'
    )

    def setUp(self):
        self.request = RequestFactory().get('/')
        self.request.user = AnonymousUser()

    def test_database_menu(self):

        template = "{% load metadata_tags %}{% databases_menu %}"

        context = RequestContext(self.request, {})
        rendered_template = Template(template).render(context)

        self.assertIn('daiquiri_data_obs', rendered_template)



from sendfile import sendfile

from django.conf import settings
from django.http import Http404, StreamingHttpResponse

from rest_framework import viewsets, mixins, filters
from rest_framework.response import Response
from rest_framework.exceptions import ValidationError, NotFound
from rest_framework.decorators import list_route, detail_route
from rest_framework.authentication import SessionAuthentication, TokenAuthentication

from daiquiri.core.viewsets import ChoicesViewSet
from daiquiri.core.permissions import HasModelPermission
from daiquiri.core.paginations import ListPagination
from daiquiri.core.utils import get_client_ip

from daiquiri.jobs.viewsets import SyncJobViewSet, AsyncJobViewSet

from .models import QueryJob, Example
from .serializers import (
    FormSerializer,
    DropdownSerializer,
    QueryJobSerializer,
    QueryJobListSerializer,
    QueryJobRetrieveSerializer,
    QueryJobCreateSerializer,
    QueryJobUpdateSerializer,
    QueryLanguageSerializer,
    ExampleSerializer,
    UserExampleSerializer,
    SyncQueryJobSerializer,
    AsyncQueryJobSerializer
)

from .permissions import HasPermission
from .utils import get_format_config, get_quota, fetch_user_database_metadata


class StatusViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    permission_classes = (HasPermission, )
    queryset = []

    def list(self, request):
        return Response([{
            'guest': not request.user.is_authenticated(),
            'queued_jobs': None,
            'size': QueryJob.objects.get_size(request.user),
            'quota': get_quota(request.user)
        }])


class FormViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    permission_classes = (HasPermission, )

    serializer_class = FormSerializer

    def get_queryset(self):
        return settings.QUERY_FORMS


class DropdownViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    permission_classes = (HasPermission, )

    serializer_class = DropdownSerializer

    def get_queryset(self):
        return settings.QUERY_DROPDOWNS


class QueryJobViewSet(viewsets.ModelViewSet):
    permission_classes = (HasPermission, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    def get_queryset(self):
        queryset = QueryJob.objects.filter_by_owner(self.request.user).exclude(phase=QueryJob.PHASE_ARCHIVED)

        # hide TAP jobs for the anonymous user
        if self.request.user.is_anonymous:
            queryset = queryset.filter(job_type=QueryJob.JOB_TYPE_INTERFACE)

        return queryset

    def get_serializer_class(self):
        if self.action == 'list':
            return QueryJobListSerializer
        elif self.action == 'retrieve' or self.action == 'abort':
            return QueryJobRetrieveSerializer
        elif self.action == 'create':
            return QueryJobCreateSerializer
        elif self.action == 'update' or self.action == 'partial_update':
            return QueryJobUpdateSerializer
        else:
            return QueryJobSerializer

    def perform_create(self, serializer):
        job = QueryJob(
            job_type=QueryJob.JOB_TYPE_INTERFACE,
            owner=(None if self.request.user.is_anonymous() else self.request.user),
            table_name=serializer.data.get('table_name'),
            query_language=serializer.data.get('query_language'),
            query=serializer.data.get('query'),
            queue=serializer.data.get('queue'),
            client_ip=get_client_ip(self.request)
        )
        job.process()
        job.save()
        job.run()

        # inject the job id into the serializers data
        serializer._data['id'] = job.id

    def perform_update(self, serializer):
        serializer.save()

    def perform_destroy(self, instance):
        instance.archive()

    @list_route(methods=['get'])
    def tables(self, request):
        return Response(fetch_user_database_metadata(request.user, self.get_queryset()))

    @detail_route(methods=['put'])
    def abort(self, request, pk=None):
        try:
            job = self.get_queryset().get(pk=pk)
            job.abort()

            serializer = QueryJobRetrieveSerializer(instance=job)
            return Response(serializer.data)
        except QueryJob.DoesNotExist:
            raise Http404

    @detail_route(methods=['get', 'put'], url_path='download/(?P<format_key>[A-Za-z0-9\-]+)', url_name='download')
    def download(self, request, pk=None, format_key=None):
        try:
            job = self.get_queryset().get(pk=pk)
        except QueryJob.DoesNotExist:
            raise NotFound

        try:
            format_config = get_format_config(format_key)
        except IndexError:
            raise ValidationError({'format': "Not supported."})

        result, file_name = job.download(format_config)

        if result.successful():
            if self.request.method == 'GET':
                return sendfile(request, file_name, attachment=True)
            else:
                return Response(result.status)

        else:
            if result.status == 'FAILURE':
                return Response(result.status, status=500)
            else:
                return Response(result.status)

    @detail_route(methods=['get'], url_path='stream/(?P<format_key>[A-Za-z0-9\-]+)', url_name='stream')
    def stream(self, request, pk=None, format_key=None):
        try:
            job = self.get_queryset().get(pk=pk)
        except QueryJob.DoesNotExist:
            raise NotFound

        try:
            format_config = get_format_config(format_key)
        except IndexError:
            raise ValidationError({'format': "Not supported."})

        return StreamingHttpResponse(job.stream(format_config), content_type=format_config['content_type'])


class ExampleViewSet(viewsets.ModelViewSet):
    permission_classes = (HasModelPermission, )
    serializer_class = ExampleSerializer
    pagination_class = ListPagination
    queryset = Example.objects.all()

    filter_backends = (
        filters.SearchFilter,
        filters.OrderingFilter
    )
    search_fields = ('name', 'description', 'query_string')

    @list_route(methods=['get'], permission_classes=(HasPermission, ))
    def user(self, request):
        examples = Example.objects.filter_by_access_level(self.request.user)
        serializer = UserExampleSerializer(examples, many=True)
        return Response(serializer.data)


class QueueViewSet(ChoicesViewSet):
    permission_classes = (HasPermission, )
    queryset = [(item['key'], item['label']) for item in settings.QUERY_QUEUES]


class QueryLanguageViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    permission_classes = (HasPermission, )
    serializer_class = QueryLanguageSerializer
    queryset = settings.QUERY_LANGUAGES


class SyncQueryJobViewSet(SyncJobViewSet):

    serializer_class = SyncQueryJobSerializer

    parameter_map = {
        'FORMAT': 'response_format',
        'TABLE_NAME': 'table_name',
        'LANG': 'query_language',
        'QUERY': 'query'
    }

    def get_queryset(self):
        return QueryJob.objects.filter_by_owner(self.request.user)


class AsyncQueryJobViewSet(AsyncJobViewSet):

    serializer_class = AsyncQueryJobSerializer

    parameter_map = {
        'FORMAT': 'response_format',
        'TABLE_NAME': 'table_name',
        'LANG': 'query_language',
        'QUEUE': 'queue',
        'QUERY': 'query'
    }

    def get_queryset(self):
        return QueryJob.objects.filter_by_owner(self.request.user)



import socket

from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):

    requires_system_checks = False
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument('database', nargs='*', help='List of additional databases.')

    def get_config(self, key):
        config = settings.DATABASES.get(key)

        if 'HOST' not in config or not config['HOST']:
            config['HOST'] = 'localhost'
            config['CLIENT'] = 'localhost'
        else:
            config['CLIENT'] = socket.gethostname()

        config['PREFIX'] = settings.QUERY_USER_DATABASE_PREFIX

        return config

    def handle(self, *args, **options):

        default = self.get_config('default')
        tap = self.get_config('tap')
        data = self.get_config('data')

        print('''-- Run the following commands on \'%(HOST)s\':

CREATE DATABASE `%(NAME)s`;
CREATE USER \'%(USER)s\'@\'%(CLIENT)s\' identified by \'%(PASSWORD)s\';
GRANT ALL PRIVILEGES ON `%(NAME)s`.* to \'%(USER)s\'@\'%(CLIENT)s\';
''' % default)

        print('''-- Run the following commands on \'%(HOST)s\':

CREATE DATABASE `%(NAME)s`;
CREATE USER \'%(USER)s\'@\'%(CLIENT)s\' identified by \'%(PASSWORD)s\';
GRANT ALL PRIVILEGES ON `%(NAME)s`.* to \'%(USER)s\'@\'%(CLIENT)s\';
''' % tap)

        print('''-- Run the following commands on \'%(HOST)s\':

CREATE USER \'%(USER)s\'@\'%(CLIENT)s\' identified by \'%(PASSWORD)s\';
GRANT ALL PRIVILEGES ON `%(PREFIX)s%%`.* to \'%(USER)s\'@\'%(CLIENT)s\';
''' % data)

        data.update({'DATABASE_NAME': tap['NAME']})
        print('GRANT SELECT ON `%(DATABASE_NAME)s`.* to \'%(USER)s\'@\'%(CLIENT)s\';' % data)

        if options['database']:
            for database_name in options['database']:
                data.update({'DATABASE_NAME': database_name})
                print('GRANT SELECT ON `%(DATABASE_NAME)s`.* to \'%(USER)s\'@\'%(CLIENT)s\';' % data)

        print('')
        print('''-- For testing, run the following commands on \'%(HOST)s\':

GRANT ALL PRIVILEGES ON `test_%(NAME)s`.* to \'%(USER)s\'@\'%(CLIENT)s\';
''' % default)

        print('''-- For testing, run the following commands on \'%(HOST)s\':

GRANT ALL PRIVILEGES ON `test_%(NAME)s`.* to \'%(USER)s\'@\'%(CLIENT)s\';
''' % tap)

        print('''-- For testing, run the following commands on \'%(HOST)s\':

GRANT ALL PRIVILEGES ON `test_`.* to \'%(USER)s\'@\'%(CLIENT)s\';
''' % data)



import os

from django.conf import settings
from django.test import TestCase

from test_generator.views import TestViewMixin


class ServeViewTestCase(TestCase):

    fixtures = (
        'auth.json',
        'files.json',
        'metadata.json',
        'jobs.json',
        'queryjobs.json'
    )

    users = (
        ('admin', 'admin'),
        ('manager', 'manager'),
        ('user', 'user'),
        ('anonymous', None),
    )


class PublicTableTests(TestViewMixin, ServeViewTestCase):

    url_names = {
        'list_view': 'serve:table'
    }

    status_map = {
        'list_view': {
            'admin': 200, 'manager': 200, 'user': 200, 'anonymous': 200
        }
    }

    def _test_list_view(self, username):
        self.assert_list_view(username, {
            'database_name': 'daiquiri_data_obs',
            'table_name': 'stars'
        })


class InternalTableTests(TestViewMixin, ServeViewTestCase):

    url_names = {
        'list_view': 'serve:table'
    }

    status_map = {
        'list_view': {
            'admin': 200, 'manager': 200, 'user': 200, 'anonymous': 404
        }
    }

    def _test_list_view(self, username):
        self.assert_list_view(username, {
            'database_name': 'daiquiri_data_sim',
            'table_name': 'particles'
        })


class UserTableTests(TestViewMixin, ServeViewTestCase):

    url_names = {
        'list_view': 'serve:table'
    }

    status_map = {
        'list_view': {
            'admin': 404, 'manager': 404, 'user': 200, 'anonymous': 404
        }
    }

    def _test_list_view(self, username):
        self.assert_list_view(username, {
            'database_name': 'daiquiri_user_user',
            'table_name': 'test'
        })


class NotFoundTableTests(TestViewMixin, ServeViewTestCase):

    url_names = {
        'list_view': 'serve:table'
    }

    status_map = {
        'list_view': {
            'admin': 404, 'manager': 404, 'user': 404, 'anonymous': 404
        }
    }

    def _test_non_existing_database_view(self, username):
        self.assert_list_view(username, {
            'database_name': 'non_existing',
            'table_name': 'stars'
        })

    def _test_non_existing_table_view(self, username):
        self.assert_list_view(username, {
            'database_name': 'daiquiri_data_obs',
            'table_name': 'non_existing'
        })

    def _test_non_existing_user_table_view(self, username):
        self.assert_list_view(username, {
            'database_name': 'daiquiri_user_user',
            'table_name': 'non_existing'
        })


class ArchiveTests(TestViewMixin, ServeViewTestCase):

    url_names = {
        'list_view': 'serve:archive'
    }

    status_map = {
        'list_view': {
            'admin': 404, 'manager': 200, 'user': 404, 'anonymous': 404
        }
    }

    def setUp(self):
        for username, password in self.users:
            try:
                os.remove(os.path.join(settings.SERVE_DOWNLOAD_DIR, username, 'images_preview.zip'))
            except OSError:
                pass

    def _test_archive_get(self, username):

        self.assert_view('list_view', 'get', 'list_view', username, kwargs={
            'database_name': 'daiquiri_data_obs',
            'table_name': 'images',
            'column_name': 'preview'
        })

        self.assert_view('list_view', 'get', 'list_view', username, kwargs={
            'database_name': 'daiquiri_data_obs',
            'table_name': 'images',
            'column_name': 'preview'
        })

    def _test_archive_put(self, username):

        self.assert_view('list_view', 'put', 'list_view', username, kwargs={
            'database_name': 'daiquiri_data_obs',
            'table_name': 'images',
            'column_name': 'preview'
        })

        self.assert_view('list_view', 'put', 'list_view', username, kwargs={
            'database_name': 'daiquiri_data_obs',
            'table_name': 'images',
            'column_name': 'preview'
        })



import subprocess

from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('queue', nargs='?', default='default', help='Name of the worker [default=default].')
        parser.add_argument('-c', type=int, default=1, help='Concurrency for the worker.')

    def handle(self, *args, **options):
        concurrency = str(options['c'])

        queue = options['queue']
        hostname = '%s_%s@%%h' % (settings.DAIQUIRI_APP, options['queue'])

        args = [
            'celery', 'worker',
            '-A', 'config',
            '-Q', queue,
            '-c', concurrency,
            '-n', hostname,
            '-l',  'info'
        ]

        subprocess.call(args)



from django.conf import settings

from daiquiri.core.utils import import_class

from .database.mysql import MySQLAdapter
from .download.mysqldump import MysqldumpAdapter

_adapter = None


class Adapter(object):

    def __init__(self):

        self.database_key = 'data'
        self.database_config = settings.DATABASES['data']

        try:
            database_adapter_class = import_class(settings.ADAPTER_DATABASE)
            self.database = database_adapter_class(self.database_key, self.database_config)
        except AttributeError:

            if self.database_config['ENGINE'] == 'django.db.backends.mysql':
                self.database = MySQLAdapter(self.database_key, self.database_config)
            else:
                raise Exception('No suitable database adapter found.')

        try:
            download_adapter_class = import_class(settings.ADAPTER_DOWNLOAD)
            self.download = download_adapter_class(self.database_key, self.database_config)
        except AttributeError:

            if self.database_config['ENGINE'] == 'django.db.backends.mysql':
                self.download = MysqldumpAdapter(self.database_key, self.database_config)
            else:
                raise Exception('No suitable download adapter found.')


def get_adapter():
    global _adapter
    if not _adapter:
        _adapter = Adapter()

    return _adapter



import base64
import csv
import io
import re
import six
import struct
import subprocess
import sys

from bitstring import BitArray

from .base import DownloadAdapter


class MysqldumpAdapter(DownloadAdapter):

    insert_pattern = re.compile('^INSERT INTO .*? VALUES \((.*?)\);')

    FORMATS = {
        'char': 'c',
        'unsignedByte': 'B',
        'short': 'h',
        'int': 'i',
        'long': 'q',
        'float': 'f',
        'double': 'd'
    }

    NULL_VALUES = {
        'char': '',
        'unsignedByte': 255,
        'short': 32767,
        'int': 2147483647,
        'long': 9223372036854775807,
        'float': float('nan'),
        'double': float('nan')
    }

    def __init__(self, database_key, database_config):
        self.args = ['mysqldump', '--compact', '--skip-extended-insert']

        if 'USER' in database_config and database_config['USER']:
            self.args.append('--user=%(USER)s' % database_config)

        if 'PASSWORD' in database_config and database_config['PASSWORD']:
            self.args.append('--password=%(PASSWORD)s' % database_config)

        if 'HOST' in database_config and database_config['HOST']:
            self.args.append('--host=%(HOST)s' % database_config)

        if 'PORT' in database_config and database_config['PORT']:
            self.args.append('--port=%(PORT)s' % database_config)

    def generate(self, format_key, database_name, table_name, metadata, status=None, empty=False):

        if format_key == 'csv':
            return self.generate_csv(database_name, table_name)

        elif format_key == 'votable':
            return self.generate_votable('TABLEDATA', database_name, table_name, metadata, status, empty)

        elif format_key == 'votable-binary':
            return self.generate_votable('BINARY', database_name, table_name, metadata, status, empty)

        elif format_key == 'votable-binary2':
            return self.generate_votable('BINARY2', database_name, table_name, metadata, status, empty)

        else:
            raise Exception('Not supported.')

    def generate_csv(self, database_name, table_name):
        args = self.args + [database_name, table_name]
        process = subprocess.Popen(args, stdout=subprocess.PIPE)

        if sys.version_info.major >= 3:
            io_class = io.StringIO
        else:
            io_class = io.BytesIO

        for line in process.stdout:
            row = self._parse_line(line)
            if row:
                f = io_class()
                csv.writer(f, quotechar='"').writerow(row)
                yield f.getvalue()

    def generate_votable(self, serialization, database_name, table_name, metadata, status, empty):
        args = self.args + [database_name, table_name]
        process = subprocess.Popen(args, stdout=subprocess.PIPE)

        fmt_list, null_value_list = self._get_fmt_list(metadata['columns'])

        yield '''<?xml version="1.0"?>
<VOTABLE version="1.3"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns="http://www.ivoa.net/xml/VOTable/v1.3"
    xmlns:stc="http://www.ivoa.net/xml/STC/v1.30">
    <RESOURCE name="%(database)s" type="results">''' % {
            'database': database_name
        }
        if status == 'OK':
            yield '''
        <INFO name="QUERY_STATUS" value="OK" />'''
        elif status == 'OVERFLOW':
            yield '''
        <INFO name="QUERY_STATUS" value="OVERFLOW" />'''

        yield '''
        <TABLE name="%(table)s">''' % {
            'table': table_name
        }

        if 'columns' in metadata:
            for i, column in enumerate(metadata['columns']):
                attrs = []
                for key in ['name', 'datatype', 'arraysize', 'unit', 'ucd', 'utype']:
                    if key in column and column[key]:
                        attrs.append('%s="%s"' % (key, column[key]))

                if serialization == 'BINARY':
                    # for binary specify excplicit null value
                    yield '''
            <FIELD %s>
                <VALUES null="%s" />
            </FIELD>''' % (' '.join(attrs), null_value_list[i])

                else:
                    # just yield the field metadata
                    yield '''
            <FIELD %s />''' % ' '.join(attrs)

        if not empty:
            yield '''
            <DATA>'''

            # write serialization specific opening tag
            if serialization == 'TABLEDATA':
                yield '''
                <TABLEDATA>'''
            elif serialization == 'BINARY':
                yield '''
                <BINARY>
                    <STREAM encoding="base64">'''
            elif serialization == 'BINARY2':
                yield '''
                <BINARY2>
                    <STREAM encoding="base64">'''

            # write rows of the table from stdout of the mysqldump process
            for line in process.stdout:
                parsed_line = self._parse_line(line)
                if parsed_line:
                    if serialization == 'TABLEDATA':
                        yield '''
                    <TR>
                        <TD>%s</TD>
                    </TR>''' % '''</TD>
                        <TD>'''.join([('' if cell == 'NULL' else cell) for cell in parsed_line])

                    elif serialization == 'BINARY':
                        values, null_mask = self._get_binary_values(parsed_line, fmt_list, null_value_list)
                        binary_string = struct.pack('>' + ''.join(fmt_list), *values)
                        yield base64.b64encode(binary_string).decode()

                    elif serialization == 'BINARY2':
                        values, null_mask = self._get_binary_values(parsed_line, fmt_list, null_value_list)
                        binary_string = BitArray(bin=null_mask).tobytes() + struct.pack('>' + ''.join(fmt_list), *values)
                        yield base64.b64encode(binary_string).decode()

            # write serialization specific closing tag
            if serialization == 'TABLEDATA':
                yield '''
                </TABLEDATA>'''
            elif serialization == 'BINARY':
                yield '''</STREAM>
                </BINARY>'''
            elif serialization == 'BINARY2':
                yield '''</STREAM>
                </BINARY2>'''

            yield '''
            </DATA>'''
        yield '''
        </TABLE>
    </RESOURCE>
</VOTABLE>
'''

    def _get_fmt_list(self, columns):
        fmt_list = []
        null_value_list = []
        for column in columns:
            if column['arraysize'] is not None:
                if column['datatype'] == 'char':
                    # this is a string!
                    fmt_list.append(str(column['arraysize']) + 's')
                else:
                    fmt_list.append(str(column['arraysize']) + self.FORMATS[column['datatype']])

                null_value_list.append('')
            else:
                fmt_list.append(self.FORMATS[column['datatype']])
                null_value_list.append(self.NULL_VALUES[column['datatype']])

        return fmt_list, null_value_list

    def _parse_line(self, line):
        insert_result = self.insert_pattern.match(line.decode())
        if insert_result:
            row = insert_result.group(1)
            reader = csv.reader([row], quotechar="'")
            return six.next(reader)
        else:
            return None

    def _get_binary_values(self, parsed_line, fmt_list, null_value_list):
        values = []
        null_mask = ''

        for i, cell in enumerate(parsed_line):
            if cell == 'NULL':
                null_mask += '1'
                values.append(null_value_list[i])

            else:
                null_mask += '0'

                if fmt_list[i] in ['B', 'h', 'i', 'q']:
                    values.append(int(cell))

                elif fmt_list[i] in ['f', 'd']:
                    values.append(float(cell))

                else:  # char
                    values.append(cell)

        return values, null_mask



import mock

from django.test import TestCase

from daiquiri.core.adapter import get_adapter
from daiquiri.jobs.tests.mixins import SyncTestMixin
from daiquiri.query.models import QueryJob, Example

adapter = get_adapter()

mock_execute = mock.Mock()

mock_fetch_stats = mock.Mock()
mock_fetch_stats.return_value = [0, 0]

mock_fetch_pid = mock.Mock()
mock_fetch_pid.return_value = 1

mock_fetch_table = mock.Mock()
mock_fetch_table.return_value = []


@mock.patch.object(adapter.database, 'execute', mock_execute)
@mock.patch.object(adapter.database, 'fetch_stats', mock_fetch_stats)
@mock.patch.object(adapter.database, 'fetch_pid', mock_fetch_pid)
@mock.patch.object(adapter.database, 'fetch_table', mock_fetch_table)
class AsyncTestCase(SyncTestMixin, TestCase):

    fixtures = (
        'auth.json',
        'metadata.json',
        'jobs.json',
        'queryjobs.json',
        'examples.json'
    )

    users = (
        ('admin', 'admin'),
        ('user', 'user'),
        ('evil', 'evil'),
        ('anonymous', None),
    )

    url_names = {
        'list': 'tap:sync-list'
    }

    jobs = QueryJob.objects.filter(owner__username='user')

    def get_parameter_for_new_jobs(self, username):
        return [{
            'LANG': example.query_language,
            'QUERY': example.query_string
        } for example in Example.objects.filter(access_level='PUBLIC')]

    def get_parameter_for_new_jobs_internal(self, username):
        return [{
            'LANG': example.query_language,
            'QUERY': example.query_string
        } for example in Example.objects.filter(access_level='INTERNAL')]



from django.db import connections


class DatabaseAdapter(object):

    def __init__(self, database_key, database_config):
        self.database_key = database_key
        self.database_config = database_config

    def connection(self):
        return connections[self.database_key]

    def execute(self, sql):
        return self.connection().cursor().execute(sql)

    def fetchone(self, sql, args=None):
        cursor = self.connection().cursor()

        if args:
            cursor.execute(sql, args)
        else:
            cursor.execute(sql)

        return cursor.fetchone()

    def fetchall(self, sql, args=None):
        cursor = self.connection().cursor()

        if args:
            cursor.execute(sql, args)
        else:
            cursor.execute(sql)

        return cursor.fetchall()

    def fetch_pid(self):
        raise NotImplementedError()

    def escape_identifier(self, identifier):
        raise NotImplementedError()

    def escape_string(self, string):
        raise NotImplementedError()

    def build_query(self, database_name, table_name, query, timeout):
        raise NotImplementedError()

    def abort_query(self, pid):
        raise NotImplementedError()

    def count_rows(self, database_name, table_name, column_names=None, filter_string=None):
        raise NotImplementedError()

    def fetch_rows(self, database_name, table_name, column_names=None, ordering=None, page=1, page_size=10, filter_string=None):
        raise NotImplementedError()

    def fetch_row(self, database_name, table_name, column_name, value):
        raise NotImplementedError()

    def create_user_database_if_not_exists(self, database_name):
        raise NotImplementedError()

    def fetch_tables(self, database_name):
        raise NotImplementedError()

    def fetch_table(self, database_name, table_name):
        raise NotImplementedError()

    def rename_table(self, database_name, table_name, new_table_name):
        raise NotImplementedError()

    def drop_table(self, database_name, table_name):
        raise NotImplementedError()

    def fetch_columns(self, database_name, table_name):
        raise NotImplementedError()

    def fetch_column(self, database_name, table_name, column_name):
        raise NotImplementedError()

    def fetch_column_names(self, database_name, table_name):
        raise NotImplementedError()



LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        },
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue'
        }
    },
    'formatters': {
        'default': {
            'format': '[%(asctime)s] %(levelname)s: %(message)s'
        },
        'name': {
            'format': '[%(asctime)s] %(levelname)s %(name)s: %(message)s'
        },
        'console': {
            'format': '[%(asctime)s] %(message)s'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        },
        'error_log': {
            'level': 'ERROR',
            'class':'logging.FileHandler',
            'filename': 'error.log',
            'formatter': 'default'
        },
        'daiquiri_log': {
            'level': 'DEBUG',
            'class':'logging.FileHandler',
            'filename': 'daiquiri.log',
            'formatter': 'name'
        },
        'sql_log': {
            'level': 'DEBUG',
            'filters': ['require_debug_true'],
            'class': 'logging.FileHandler',
            'filename': 'sql.log',
            'formatter': 'default'
        },
        'rules_log': {
            'level': 'DEBUG',
            'filters': ['require_debug_true'],
            'class': 'logging.FileHandler',
            'filename': 'rules.log',
            'formatter': 'default'
        },
        'console': {
            'level': 'DEBUG',
            'filters': ['require_debug_true'],
            'class': 'logging.StreamHandler',
            'formatter': 'console'
        }
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'django.request': {
            'handlers': ['mail_admins', 'error_log'],
            'level': 'ERROR',
            'propagate': True
        },
        'django.db.backends': {
            'handlers': ['sql_log'],
            'level': 'DEBUG',
            'propagate': False
        },
        'daiquiri': {
            'handlers': ['daiquiri_log'],
            'level': 'DEBUG',
            'propagate': False
        },
        'rules': {
            'handlers': ['rules_log'],
            'level': 'DEBUG',
            'propagate': False
        }
    }
}



from rest_framework.routers import Route, SimpleRouter


class JobRouter(SimpleRouter):
    routes = [
        Route(
            url=r'^{prefix}$',
            mapping={'get': 'list_jobs', 'post': 'create_job'},
            name='{basename}-list',
            initkwargs={'suffix': 'List'}
        ),
        Route(
            url=r'^{prefix}/{lookup}$',
            mapping={'get': 'retrieve_job', 'post': 'update_job', 'delete': 'destroy_job'},
            name='{basename}-detail',
            initkwargs={'suffix': 'Detail'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/results$',
            mapping={'get': 'get_results'},
            name='{basename}-results',
            initkwargs={'suffix': 'Results'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/results/result$',
            mapping={'get': 'get_result'},
            name='{basename}-result',
            initkwargs={'suffix': 'Result'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/parameters$',
            mapping={'get': 'get_parameters'},
            name='{basename}-parameters',
            initkwargs={'suffix': 'Parameters'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/destruction$',
            mapping={'get': 'get_destruction', 'post': 'set_destruction'},
            name='{basename}-destruction',
            initkwargs={'suffix': 'Destruction'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/executionduration$',
            mapping={'get': 'get_executionduration', 'post': 'set_executionduration'},
            name='{basename}-executionduration',
            initkwargs={'suffix': 'Executionduration'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/phase$',
            mapping={'get': 'get_phase', 'post': 'set_phase'},
            name='{basename}-phase',
            initkwargs={'suffix': 'Phase'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/error$',
            mapping={'get': 'get_error'},
            name='{basename}-error',
            initkwargs={'suffix': 'Error'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/quote$',
            mapping={'get': 'get_quote'},
            name='{basename}-quote',
            initkwargs={'suffix': 'Quote'}
        ),
        Route(
            url=r'^{prefix}/{lookup}/owner$',
            mapping={'get': 'get_owner'},
            name='{basename}-owner',
            initkwargs={'suffix': 'Owner'}
        ),
    ]



from django.contrib.auth.models import Group

from rest_framework import serializers

from ..models import Database, Table, Column, Function


class GroupSerializer(serializers.ModelSerializer):

    class Meta:
        model = Group
        fields = ('id', 'name')


class FunctionSerializer(serializers.ModelSerializer):

    label = serializers.CharField(source='__str__', read_only=True)

    class Meta:
        model = Function
        fields = '__all__'


class ColumnSerializer(serializers.ModelSerializer):

    label = serializers.CharField(source='__str__', read_only=True)

    class Meta:
        model = Column
        fields = '__all__'


class TableSerializer(serializers.ModelSerializer):

    label = serializers.CharField(source='__str__', read_only=True)

    class Meta:
        model = Table
        fields = '__all__'


class DatabaseSerializer(serializers.ModelSerializer):

    label = serializers.CharField(source='__str__', read_only=True)

    class Meta:
        model = Database
        fields = '__all__'



from django.conf import settings
from django.http import HttpResponse
from django.urls import reverse

from daiquiri.metadata.models import Database
from daiquiri.query.models import Example

from .serializers import ExampleSerializer, SchemaSerializer
from .renderers import (
    ExampleRenderer,
    AvailabilityRenderer,
    CapabilitiesRenderer,
    TablesetRenderer
)


def examples(request):
    queryset = Example.objects.filter_by_access_level(request.user)
    serializer = ExampleSerializer(queryset, many=True)
    return HttpResponse(ExampleRenderer().render(serializer.data), content_type="application/xml")


def availability(request):
    data = {
        'available': 'true',
        'note': 'service is accepting queries'
    }
    return HttpResponse(AvailabilityRenderer().render(data), content_type="application/xml")


def capabilities(request):
    data = [
        {
            'schemaID': 'ivo://ivoa.net/std/TAP',
            'interface': {
                'attrs': {
                    'xsi:type': 'vs:ParamHTTP',
                    'role': 'std'
                },
                'accessURL': {
                    'attrs': {},
                    'text': request.build_absolute_uri(reverse('tap:root'))
                }
            },
            'languages': [{
                'name': language['key'],
                'version': language['version'],
                'description': language['description'],
            } for language in settings.QUERY_LANGUAGES]

        },
        {
            'schemaID': 'ivo://ivoa.net/std/TAP#async-1.1',
            'interface': {
                'attrs': {
                    'xsi:type': 'vs:ParamHTTP',
                    'role': 'std',
                    'version': '1.1'
                },
                'accessURL': {
                    'attrs': {
                        'use': 'base'
                    },
                    'text': request.build_absolute_uri(reverse('tap:async-list'))
                }
            }
        },
        {
            'schemaID': 'ivo://ivoa.net/std/TAP#sync-1.1',
            'interface': {
            'attrs': {
                'xsi:type': 'vs:ParamHTTP',
                'role': 'std',
                'version': '1.1'
                },
                'accessURL': {
                    'attrs': {
                        'use': 'base'
                    },
                    'text': request.build_absolute_uri(reverse('tap:sync-list'))
                }
            }
        },
        {
            'schemaID': 'ivo://ivoa.net/std/VOSI#capabilities',
            'interface': {
            'attrs': {
                'xsi:type': 'vs:ParamHTTP',
                },
                'accessURL': {
                    'attrs': {
                        'use': 'full'
                    },
                    'text': request.build_absolute_uri(reverse('tap:capabilities'))
                }
            }
        },
        {
            'schemaID': 'ivo://ivoa.net/std/VOSI#tables',
            'interface': {
            'attrs': {
                'xsi:type': 'vs:ParamHTTP',
                },
                'accessURL': {
                    'attrs': {
                        'use': 'full'
                    },
                    'text': request.build_absolute_uri(reverse('tap:tables'))
                }
            }
        },
        {
            'schemaID': 'ivo://ivoa.net/std/DALI#examples',
            'interface': {
            'attrs': {
                'xsi:type': 'vr:WebBrowser',
                },
                'accessURL': {
                    'attrs': {
                        'use': 'full'
                    },
                    'text': request.build_absolute_uri(reverse('tap:examples'))
                }
            }
        }
    ]

    return HttpResponse(CapabilitiesRenderer().render(data), content_type="application/xml")

def tables(request):
    queryset = Database.objects.filter_by_access_level(request.user)
    serializer = SchemaSerializer(queryset, context={'request': request}, many=True)
    return HttpResponse(TablesetRenderer().render(serializer.data), content_type="application/xml")



from django.conf.urls import url, include

from rest_framework import routers

from .views import contact, MessagesView
from .viewsets import ContactMessageViewSet, StatusViewSet

router = routers.DefaultRouter()
router.register(r'messages', ContactMessageViewSet, base_name='message')
router.register(r'status', StatusViewSet, base_name='status')

urlpatterns = [
    url(r'^$', contact, name='contact'),
    url(r'^messages/', MessagesView.as_view(), name='messages'),

    # rest api
    url(r'^api/', include(router.urls)),
]



import sys

from django.core.management.base import BaseCommand

from daiquiri.core.adapter import get_adapter
from daiquiri.metadata.models import Table


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('database', help='the database for the table to dump')
        parser.add_argument('table', help='the table to dump')
        parser.add_argument('format', help='the format for the dump')


    def handle(self, *args, **options):
        table = Table.objects.filter(database__name=options['database']).get(name=options['table'])

        for line in get_adapter().download.generate(options['format'], options['database'], options['table'], {
                'columns': table.columns.values()
            }):
            sys.stdout.write(line)



from __future__ import absolute_import, unicode_literals

import subprocess

from django.conf import settings

from celery import shared_task

from daiquiri.core.tasks import Task


@shared_task(base=Task)
def create_wordpress_user(username, email, first_name, last_name):
    return subprocess.check_output([
        settings.WORDPRESS_CLI,
        'user',
        'create',
        username,
        email,
        '--first_name=%s' % first_name,
        '--last_name=%s' % last_name,
        '--path=%s' % settings.WORDPRESS_PATH
    ])


@shared_task(base=Task)
def update_wordpress_user(username, email, first_name, last_name):
    return subprocess.check_output([
        settings.WORDPRESS_CLI,
        'user',
        'update',
        username,
        '--user_email=%s' % email,
        '--first_name=%s' % first_name,
        '--last_name=%s' % last_name,
        '--path=%s' % settings.WORDPRESS_PATH
    ])


@shared_task(base=Task)
def update_wordpress_role(username, role):
    return subprocess.check_output([
        settings.WORDPRESS_CLI,
        'user',
        'set-role',
        username,
        role,
        '--path=%s' % settings.WORDPRESS_PATH
    ])



from rest_framework import mixins, viewsets

from .serializers import ChoicesSerializer


class ChoicesViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = ChoicesSerializer



from django.db import models

from .constants import ACCESS_LEVEL_PUBLIC, ACCESS_LEVEL_INTERNAL


class AccessLevelManager(models.Manager):

    def filter_by_access_level(self, user):
        if not user or user.is_anonymous():
            return self.get_queryset().filter(access_level=ACCESS_LEVEL_PUBLIC)
        else:
            q = models.Q(access_level=ACCESS_LEVEL_PUBLIC) | \
                models.Q(access_level=ACCESS_LEVEL_INTERNAL) | \
                models.Q(groups__in=user.groups.all())
            return self.get_queryset().filter(q)

    def filter_by_metadata_access_level(self, user):
        if not user or user.is_anonymous():
            return self.get_queryset().filter(metadata_access_level=ACCESS_LEVEL_PUBLIC)
        else:
            q = models.Q(metadata_access_level=ACCESS_LEVEL_PUBLIC) | \
                models.Q(metadata_access_level=ACCESS_LEVEL_INTERNAL) | \
                models.Q(groups__in=user.groups.all())
            return self.get_queryset().filter(q)



from django.db import models

from daiquiri.core.managers import AccessLevelManager


class QueryJobManager(models.Manager):

    def filter_by_owner(self, user):
        if not user or user.is_anonymous():
            return self.get_queryset().filter(owner=None)
        else:
            return self.get_queryset().filter(owner=user)

    def get_size(self, user):
        # get the size of all the tables of this user
        return self.filter_by_owner(user).exclude(phase=self.model.PHASE_ARCHIVED).aggregate(models.Sum('size'))['size__sum'] or 0


class ExampleManager(AccessLevelManager):

    pass



import os
import sys

from django.conf import settings
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from daiquiri.core.adapter import get_adapter
from daiquiri.core.utils import human2bytes
from daiquiri.metadata.models import Database, Table, Column, Function


def get_format_config(format_key):
    return [f for f in settings.QUERY_DOWNLOAD_FORMATS if f['key'] == format_key][0]


def get_default_table_name():
    return now().strftime("%Y-%m-%d-%H-%M-%S")


def get_default_queue():
    return settings.QUERY_QUEUES[0]['key']


def get_query_language_choices():
    return [('%(key)s-%(version)s' % item, item['label']) for item in settings.QUERY_LANGUAGES]


def get_queue_choices():
    return [(item['key'], item['label']) for item in settings.QUERY_QUEUES]


def get_tap_schema_name():
    tap_config = settings.DATABASES.get('tap')
    if tap_config:
        return tap_config['NAME']
    else:
        return None


def get_user_database_name(user):
    if not user or user.is_anonymous():
        username = 'anonymous'
    else:
        username = user.username

    return settings.QUERY_USER_DATABASE_PREFIX + username


def get_quota(user):
    if not user or user.is_anonymous():
        anonymous_quota = human2bytes(settings.QUERY_QUOTA.get('anonymous'))
        return anonymous_quota if anonymous_quota else 0

    else:
        user_quota = human2bytes(settings.QUERY_QUOTA.get('user'))
        quota = user_quota if user_quota else 0

        # apply quota for user
        user_quotas = settings.QUERY_QUOTA.get('users')
        if user_quotas:
            user_quota = human2bytes(user_quotas.get(user.username))
            if user_quota:
                quota = user_quota if user_quota > quota else quota

        # apply quota for group
        group_quotas = settings.QUERY_QUOTA.get('groups')
        if group_quotas:
            for group in user.groups.all():
                group_quota = human2bytes(group_quotas.get(group.name))
                if group_quota:
                    quota = group_quota if group_quota > quota else quota

    return quota


def get_download_file_name(user, table_name, format_config):
    if not user or user.is_anonymous():
        username = 'anonymous'
    else:
        username = user.username

    directory_name = os.path.join(settings.QUERY_DOWNLOAD_DIR, username)
    return os.path.join(directory_name, table_name + '.' + format_config['extension'])


def fetch_user_database_metadata(user, jobs):

    database_name = get_user_database_name(user)

    database = {
        'order': sys.maxsize,
        'name': database_name,
        'query_strings': [database_name],
        'description': _('Your personal database'),
        'tables': []
    }

    for job in jobs:
        if job.phase == job.PHASE_COMPLETED:
            table = {
                'name': job.table_name,
                'query_strings': [database_name, job.table_name],
                'columns': job.metadata['columns']
            }

            for column in table['columns']:
                column['query_strings'] = [column['name']]

            database['tables'].append(table)

    return [database]


def get_asterisk_columns(display_column):
    database_name, table_name, _ = display_column[1]
    column_names = get_adapter().database.fetch_column_names(database_name, table_name)
    return [(column_name, (database_name, table_name, column_name)) for column_name in column_names]


def check_permissions(user, keywords, columns, functions):
    messages = []

    # check keywords against whitelist
    for keywords in keywords:
        pass

    # check permissions on databases/tables/columns
    for column in columns:
        database_name, table_name, column_name = column

        # check permission on database
        if database_name is None:
            continue
        elif database_name == get_tap_schema_name():
            continue
        elif database_name == get_user_database_name(user):
            continue
        else:
            try:
                database = Database.objects.filter_by_access_level(user).get(name=database_name)
            except Database.DoesNotExist:
                messages.append(_('Database %s not found.') % database_name)
                continue

        # check permission on table
        if table_name is None:
            continue
        else:
            try:
                table = Table.objects.filter_by_access_level(user).filter(database=database).get(name=table_name)
            except Table.DoesNotExist:
                messages.append(_('Table %s not found.') % table_name)
                continue

        # check permission on column
        if column_name is None:
            continue
        elif column_name == '*':
            columns = Column.objects.filter_by_access_level(user).filter(table=table)
            actual_columns = get_adapter().database.fetch_columns(database_name, table_name)

            column_names_set = set([column.name for column in columns])
            actual_column_names_set = set([column['name'] for column in actual_columns])

            if column_names_set != actual_column_names_set:
                messages.append(_('The asterisk (*) is not allowed for this table.'))
                continue
        else:
            try:
                column = Column.objects.filter_by_access_level(user).filter(table=table).get(name=column_name)
            except Column.DoesNotExist:
                messages.append(_('Column %s not found.') % column_name)
                continue



    # check permissions on functions
    for function_name in functions:
        if function_name.upper() in get_adapter().database.FUNCTIONS:
            continue
        else:
            # check permission on function
            function = Function.objects.filter_by_access_level(user).get(name=function_name)
            if not function:
                messages.append(_('Function %s not found.') % function_name)
                continue

    # return the error stack
    return list(set(messages))



from django.test import TestCase

from test_generator.views import TestListViewMixin


class CoreViewTestCase(TestCase):

    fixtures = (
        'auth.json',
    )

    users = (
        ('admin', 'admin'),
        ('user', 'user'),
        ('anonymous', None),
    )

    status_map = {
        'list_view': {
            'admin': 200, 'user': 200, 'anonymous': 200
        }
    }


class HomeTests(TestListViewMixin, CoreViewTestCase):

    url_names = {
        'list_view': 'home'
    }



import iso8601

from lxml import objectify

from django.core.urlresolvers import reverse
from django.db.models import Q
from django.utils.http import urlencode

from test_generator.core import TestMixin


class SyncTestMixin(TestMixin):

    def _test_get_job_list(self, username):
        '''
        GET /{jobs} returns 404.
        '''
        url = reverse(self.url_names['list'])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 405)

    def _test_post_job_list_create(self, username):
        '''
        POST /{jobs} with an application/x-www-form-urlencoded set of KEY=VALUE
        creates a job with these parameters and redirects to /{jobs}/{job-id}
        as 303.
        '''
        for new_job in self.get_parameter_for_new_jobs(username):
            url = reverse(self.url_names['list'])
            response = self.client.post(url, urlencode(new_job), content_type='application/x-www-form-urlencoded')
            self.assertEqual(response.status_code, 303, msg=(
                ('username', username),
                ('url', url),
                ('data', new_job),
                ('status_code', response.status_code),
                ('content', response.content)
            ))

    def _test_post_job_list_create_internal(self, username):
        '''
        POST /{jobs} with an application/x-www-form-urlencoded set of KEY=VALUE
        creates a job with these parameters and redirects to /{jobs}/{job-id}
        as 303.
        '''
        for new_job in self.get_parameter_for_new_jobs_internal(username):
            url = reverse(self.url_names['list'])
            response = self.client.post(url, urlencode(new_job), content_type='application/x-www-form-urlencoded')
            self.assertEqual(response.status_code, 400 if username == 'anonymous' else 303, msg=(
                ('username', username),
                ('url', url),
                ('data', new_job),
                ('status_code', response.status_code),
                ('content', response.content)
            ))


class AsyncTestMixin(TestMixin):

    uws_ns = '{http://www.ivoa.net/xml/UWS/v1.0}'

    def _test_get_job_list_xml(self, username):
        '''
        GET /{jobs} returns the job list as <uws:jobs> xml element. The archived
        jobs are not returned.
        '''
        url = reverse(self.url_names['list'])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        if username == 'user':
            root = objectify.fromstring(response.content)
            self.assertEqual(root.tag, self.uws_ns + 'jobs')
            self.assertEqual(root.jobref.tag, self.uws_ns + 'jobref')
            self.assertEqual(len(root.jobref), len(self.jobs))

    def _test_get_job_list_xml_phase(self, username):
        '''
        GET /{jobs}?PHASE=<phase> returns the filtered joblist as <jobs>
        element.
        '''
        url = reverse(self.url_names['list']) + '?PHASE=PENDING&PHASE=ARCHIVED'
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        if username == 'user':
            root = objectify.fromstring(response.content)
            self.assertEqual(len(root.jobref), len(self.jobs.filter(Q(phase='PENDING')|Q(phase='ARCHIVED'))))

    def _test_get_job_list_xml_after(self, username):
        '''
        GET /{jobs}?AFTER=2014-09-10T10:01:02.000 returns jobs with startTimes
        after the given [std:iso8601] time in UTC. The archived jobs are not
        returned.
        '''
        url = reverse(self.url_names['list']) + '?AFTER=2017-01-01T01:00:00Z'
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        if username == 'user':
            root = objectify.fromstring(response.content)
            self.assertEqual(len(root.jobref), len(self.jobs.filter(creation_time__gte='2017-01-01T01:00:00Z')))

    def _test_get_job_list_xml_last(self, username):
        '''
        GET /{jobs}?LAST=100 returns the given number of most recent jobs
        ordered by ascending startTimes. The archived jobs are not returned.
        '''
        url = reverse(self.url_names['list']) + '?LAST=3'
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        if username == 'user':
            root = objectify.fromstring(response.content)
            self.assertEqual(len(root.jobref), 3)

    def _test_post_job_list_create(self, username):
        '''
        POST /{jobs} with an application/x-www-form-urlencoded set of KEY=VALUE
        creates a job with these parameters and redirects to /{jobs}/{job-id}
        as 303.
        '''
        for new_job in self.get_parameter_for_new_jobs(username):
            url = reverse(self.url_names['list'])
            response = self.client.post(url, urlencode(new_job), content_type='application/x-www-form-urlencoded')
            self.assertEqual(response.status_code, 303, msg=(
                ('username', username),
                ('url', url),
                ('data', new_job),
                ('status_code', response.status_code),
                ('content', response.content)
            ))

            response = self.client.get(response['Location'] + '/phase')
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content.decode(), 'PENDING')

    def _test_post_job_list_create_internal(self, username):
        '''
        POST /{jobs} with an application/x-www-form-urlencoded set of KEY=VALUE
        creates a job with these parameters and redirects to /{jobs}/{job-id}
        as 303.
        '''
        for new_job in self.get_parameter_for_new_jobs_internal(username):
            url = reverse(self.url_names['list'])
            response = self.client.post(url, urlencode(new_job), content_type='application/x-www-form-urlencoded')
            self.assertEqual(response.status_code, 400 if username == 'anonymous' else 303, msg=(
                ('username', username),
                ('url', url),
                ('data', new_job),
                ('status_code', response.status_code),
                ('content', response.content)
            ))

    def _test_post_job_list_create_run(self, username):
        '''
        POST /{jobs} with an application/x-www-form-urlencoded set of
        KEY=VALUE and additionally PHASE=RUN to an non-existing {job-id} creates
        a job with these parameters and runs it.
        '''
        for new_job in self.get_parameter_for_new_jobs(username):
            new_job.update({'PHASE': 'RUN'})

            url = reverse(self.url_names['list'])
            response = self.client.post(url, urlencode(new_job), content_type='application/x-www-form-urlencoded')
            self.assertEqual(response.status_code, 303)

            response = self.client.get(response['Location'] + '/phase')
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content.decode(), 'COMPLETED')

    def _test_get_job_detail(self, username):
        '''
        GET /{jobs}/{job-id} returns a job as <uws:job> xml element.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['detail'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)

                root = objectify.fromstring(response.content)
                self.assertEqual(root.tag, self.uws_ns + 'job')
            else:
                self.assertEqual(response.status_code, 404)

    def _test_get_job_results(self, username):
        '''
        GET /{jobs}/{job-id}/results returns any results of the job {job-id} as
        <uws:results> xml element.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['results'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)

                root = objectify.fromstring(response.content)
                self.assertEqual(root.tag, self.uws_ns + 'results')
            else:
                self.assertEqual(response.status_code, 404)


    def _test_get_job_result(self, username):
        '''
        GET /{jobs}/{job-id}/results/result returns a 303 to the stream url for this job.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['result'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                if job.phase == 'COMPLETED':
                    self.assertRedirects(response, job.result, status_code=303, target_status_code=200)
                else:
                    self.assertRedirects(response, job.result, status_code=303, target_status_code=400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_get_job_parameters(self, username):
        '''
        GET /{jobs}/{job-id}/parameters returns any parameters for the job
        {job-id} as <uws:parameters> xml element.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['parameters'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)

                root = objectify.fromstring(response.content)
                self.assertEqual(root.tag, self.uws_ns + 'parameters')
            else:
                self.assertEqual(response.status_code, 404)

    def _test_get_job_error(self, username):
        '''
        GET /{jobs}/{job-id}/error returns any error message associated with
        {job-id} as text.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['error'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)
            else:
                self.assertEqual(response.status_code, 404)

    def _test_get_job_delete(self, username):
        '''
        DELETE /{jobs}/{job-id} sets the job phase to ARCHIVED and deletes the
        results and redirects to /{jobs}/{job-id} as 303.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['detail'], kwargs={'pk': job.pk})
            response = self.client.delete(url)

            if username == 'user':
                redirect_url = 'http://testserver' + reverse(self.url_names['detail'], kwargs={'pk': job.pk})
                self.assertRedirects(response, redirect_url, status_code=303)
                self.assertEqual(self.jobs.get(pk=job.pk).phase, 'ARCHIVED')
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_delete(self, username):
        '''
        POST /{jobs}/{job-id} with ACTION=DELETE sets the job phase to ARCHIVED
        and deletes the results and redirects to /{jobs}/{job-id} as 303.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['detail'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'ACTION': 'DELETE'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                redirect_url = 'http://testserver' + reverse(self.url_names['detail'], kwargs={'pk': job.pk})
                self.assertRedirects(response, redirect_url, status_code=303)
                self.assertEqual(self.jobs.get(pk=job.pk).phase, 'ARCHIVED')
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_invalid(self, username):
        '''
        POST /{jobs}/{job-id} with invalid ACTION returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['detail'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'ACTION': 'not_valid'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_missing(self, username):
        '''
        POST /{jobs}/{job-id} without ACTION returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['detail'], kwargs={'pk': job.pk})
            response = self.client.post(url, content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_get_job_destruction(self, username):
        '''
        GET /{jobs}/{job-id}/destruction returns the destruction instant for
        {job-id} as [std:iso8601].
        '''
        for job in self.jobs:
            url = reverse(self.url_names['destruction'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)

                if job.destruction_time:
                    destruction_time = iso8601.parse_date(response.content.decode())
                    self.assertEqual(destruction_time, job.destruction_time)
                else:
                    self.assertEqual(response.content.decode(), '')
            else:
                self.assertEqual(response.status_code, 404)

    def _test_post_job_destruction(self, username):
        '''
        POST /{jobs}/{job-id}/destruction with DESTRUCTION={std:iso8601}
        (application/x-www-form-urlencoded) sets the destruction instant for
        {job-id} and redirects to /{jobs}/{job-id} as 303.
        '''
        destruction_time = '2016-01-01T00:00:00'

        for job in self.jobs:
            url = reverse(self.url_names['destruction'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'DESTRUCTION': destruction_time}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                redirect_url = 'http://testserver' + reverse(self.url_names['detail'], kwargs={'pk': job.pk})
                self.assertRedirects(response, redirect_url, status_code=303)
                self.assertEqual(
                    self.jobs.get(pk=job.pk).destruction_time,
                    iso8601.parse_date('2016-01-01T00:00:00')
                )
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_destruction_invalid(self, username):
        '''
        POST /{jobs}/{job-id}/destruction with invalid DESTRUCTION returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['destruction'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'DESTRUCTION': 'not_a_date'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_destruction_missing(self, username):
        '''
        POST /{jobs}/{job-id}/destruction without DESTRUCTION returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['destruction'], kwargs={'pk': job.pk})
            response = self.client.post(url, content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_get_job_executionduration(self, username):
        '''
        GET /{jobs}/{job-id}/executionduration returns the maximum execution
        duration of {job-id} as integer.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['executionduration'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)
                self.assertEqual(int(response.content), job.execution_duration)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_executionduration(self, username):
        '''
        POST /{jobs}/{job-id}/executionduration with EXECUTIONDURATION={int}
        sets the maximum execution duration of {job-id} and redirects as to
        /{jobs}/{job-id} 303.
        '''
        execution_duration = 60

        for job in self.jobs:
            url = reverse(self.url_names['executionduration'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'EXECUTIONDURATION': 60}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                redirect_url = 'http://testserver' + reverse(self.url_names['detail'], kwargs={'pk': job.pk})
                self.assertRedirects(response, redirect_url, status_code=303)
                self.assertEqual(self.jobs.get(pk=job.pk).execution_duration, execution_duration)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_executionduration_invalid(self, username):
        '''
        POST /{jobs}/{job-id}/executionduration with invalid EXECUTIONDURATION returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['executionduration'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'EXECUTIONDURATION': 'not_an_integer'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_executionduration_missing(self, username):
        '''
        POST /{jobs}/{job-id}/executionduration without EXECUTIONDURATION returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['executionduration'], kwargs={'pk': job.pk})
            response = self.client.post(url, content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_get_job_phase(self, username):
        '''
        GET /{jobs}/{job-id}/phase returns the phase of job {job-id} as one of
        the fixed strings.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['phase'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content, job.phase.encode())
            else:
                self.assertEqual(response.status_code, 404)

    def _test_post_job_phase_run(self, username):
        '''
        POST /{jobs}/{job-id}/phase with PHASE=RUN runs the job {job-id} and
        redirects as to /{jobs}/{job-id} 303.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['phase'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'PHASE': 'RUN'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                if job.phase in ['PENDING']:
                    redirect_url = 'http://testserver' + reverse(self.url_names['detail'], kwargs={'pk': job.pk})
                    self.assertRedirects(response, redirect_url, status_code=303)
                    self.assertEqual(self.jobs.get(pk=job.pk).phase, 'COMPLETED')
                else:
                    self.assertEqual(response.status_code, 400)

            else:
                self.assertEqual(response.status_code, 404)

    def _test_post_job_phase_abort(self, username):
        '''
        POST /{jobs}/{job-id}/phase with PHASE=ABORT aborts the job {job-id} and
        redirects as to /{jobs}/{job-id} 303.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['phase'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'PHASE': 'ABORT'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                redirect_url = 'http://testserver' + reverse(self.url_names['detail'], kwargs={'pk': job.pk})
                self.assertRedirects(response, redirect_url, status_code=303)

                if job.phase in ['PENDING', 'QUEUED', 'EXECUTING']:
                    self.assertEqual(self.jobs.get(pk=job.pk).phase, 'ABORTED')
                else:
                    self.assertEqual(self.jobs.get(pk=job.pk).phase, job.phase)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_phase_invalid(self, username):
        '''
        POST /{jobs}/{job-id}/phase with invalid PHASE returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['phase'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'PHASE': 'invalid'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_phase_missing(self, username):
        '''
        POST /{jobs}/{job-id}/phase without PHASE returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['phase'], kwargs={'pk': job.pk})
            response = self.client.post(url, content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_post_job_phase_unsupported(self, username):
        '''
        POST /{jobs}/{job-id}/phase with PHASE=unsupported returns 400.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['phase'], kwargs={'pk': job.pk})
            response = self.client.post(url, urlencode({'PHASE': 'unsupported'}), content_type='application/x-www-form-urlencoded')

            if username == 'user':
                self.assertEqual(response.status_code, 400)
            else:
                self.assertEqual(response.status_code, 404)


    def _test_get_job_quote(self, username):
        '''
        GET /{jobs}/{job-id}/quote returns the quote for {job-id} as [std:iso8601].
        '''
        for job in self.jobs:
            url = reverse(self.url_names['quote'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content.decode(), '')
            else:
                self.assertEqual(response.status_code, 404)

    def _test_get_job_owner(self, username):
        '''
        GET /{jobs}/{job-id}/owner returns the owner of the job {job-id} as an
        appropriate identifier.
        '''
        for job in self.jobs:
            url = reverse(self.url_names['owner'], kwargs={'pk': job.pk})
            response = self.client.get(url)

            if username == 'user':
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content.decode(), job.owner.username)
            else:
                self.assertEqual(response.status_code, 404)



class TapRouter(object):

    def db_for_read(self, model, **hints):
        if model._meta.app_label == 'daiquiri_tap':
            return 'tap'
        return None

    def db_for_write(self, model, **hints):
        if model._meta.app_label == 'daiquiri_tap':
            return 'tap'
        return None

    def allow_relation(self, obj1, obj2, **hints):
        if obj1._meta.app_label == 'tap' or \
           obj2._meta.app_label == 'tap':
           return True
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if db == 'tap':
            if app_label == 'daiquiri_tap':
                return True
            else:
                return False
        else:
            if app_label == 'daiquiri_tap':
                return False
            else:
                # pass on to the next router
                return None



from django import template

from ..models import Database

register = template.Library()


@register.inclusion_tag('metadata/tags/databases_menu.html', takes_context=True)
def databases_menu(context):

    databases = Database.objects.filter_by_metadata_access_level(context.request.user)

    context['databases'] = []
    for database in databases:
        context['databases'].append({
            'name': database.name,
            'tables': [table.name for table in database.tables.filter_by_metadata_access_level(context.request.user)]
        })

    return context



from django import forms
from django.conf import settings
from django.contrib.auth.models import User
from django.utils.translation import ugettext as _

from .models import Profile


class UserForm(forms.ModelForm):
    class Meta:
        model = User
        fields = ('first_name', 'last_name')
        widgets = {
            'first_name': forms.TextInput(attrs={'placeholder': _('First name')}),
            'last_name': forms.TextInput(attrs={'placeholder': _('Last name')})
        }


class ProfileForm(forms.ModelForm):

    class Meta:
        model = Profile
        fields = ()

    def __init__(self, *args, **kwargs):
        super(ProfileForm, self).__init__(*args, **kwargs)

        # add a field for each detail key
        for detail_key in settings.AUTH_DETAIL_KEYS:

            choices = [(option['id'], option['label']) for option in detail_key['options']]

            if detail_key['data_type'] == 'text':
                field = forms.CharField(widget=forms.TextInput(attrs={'placeholder': detail_key['label']}))
            elif detail_key['data_type'] == 'textarea':
                field = forms.CharField(widget=forms.Textarea(attrs={'placeholder': detail_key['label']}))
            elif detail_key['data_type'] == 'select':
                field = forms.ChoiceField(choices=choices)
            elif detail_key['data_type'] == 'radio':
                field = forms.ChoiceField(choices=choices, widget=forms.RadioSelect)
            elif detail_key['data_type'] == 'multiselect':
                field = forms.MultipleChoiceField(choices=choices)
            elif detail_key['data_type'] == 'checkbox':
                field = forms.MultipleChoiceField(choices=choices, widget=forms.CheckboxSelectMultiple)
            else:
                raise Exception('Unknown detail key data type.')

            if 'label' in detail_key:
                field.label = detail_key['label']

            if 'required' in detail_key:
                field.required = detail_key['required']

            if 'help_text' in detail_key:
                field.help_text = detail_key['help_text']

            if self.instance.details and detail_key['key'] in self.instance.details:
                field.initial = self.instance.details[detail_key['key']]

            self.fields[detail_key['key']] = field

    def save(self, *args, **kwargs):
        # create an empty details dict if it does not exist
        if not self.instance.details:
            self.instance.details = {}

        # store the form date for each detail key
        for detail_key in settings.AUTH_DETAIL_KEYS:
            self.instance.details[detail_key['key']] = self.cleaned_data[detail_key['key']]

        return super(ProfileForm, self).save(*args, **kwargs)


class SignupForm(ProfileForm):

    first_name = forms.CharField(max_length=30, label=_('First name'), widget=forms.TextInput(attrs={'placeholder': _('First name')}))
    last_name = forms.CharField(max_length=30, label=_('Last name'), widget=forms.TextInput(attrs={'placeholder': _('Last name')}))

    def signup(self, request, user):
        # create an empty details dict
        user.profile.details = {}

        # store the form date for each detail key
        for detail_key in settings.AUTH_DETAIL_KEYS:
            user.profile.details[detail_key['key']] = self.cleaned_data[detail_key['key']]

        # save the profile model
        user.profile.save()



from __future__ import absolute_import, unicode_literals

import os
import logging

from celery import shared_task

from django.db.utils import OperationalError, ProgrammingError, InternalError
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from daiquiri.core.tasks import Task


class RunQueryTask(Task):

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        super(RunQueryTask, self).on_failure(exc, task_id, args, kwargs, einfo)

        # always import daiquiri packages inside the task
        from daiquiri.query.models import QueryJob

        # get logger
        logger = logging.getLogger(__name__)

        # get job_id from the original task args
        job_id = args[0]

        # log raised exception
        logger.error('run_query %s raised an exception (%s)' % (job_id, exc))

        # set phase and error_summary of the crashed job
        job = QueryJob.objects.get(pk=job_id)
        job.phase = job.PHASE_ERROR
        job.error_summary = str(_('There has been an server error with your job.'))
        job.save()


@shared_task(base=RunQueryTask)
def run_query(job_id):
    # always import daiquiri packages inside the task
    from daiquiri.core.adapter import get_adapter
    from daiquiri.metadata.models import Column
    from daiquiri.query.models import QueryJob
    from daiquiri.query.utils import get_quota

    # get logger
    logger = logging.getLogger(__name__)

    # get the job object from the database
    job = QueryJob.objects.get(pk=job_id)

    # get the adapter with the database specific functions
    adapter = get_adapter()

    # create the database of the user if it not already exists
    try:
        adapter.database.create_user_database_if_not_exists(job.database_name)
    except OperationalError as e:
        job.phase = job.PHASE_ERROR
        job.error_summary = str(e)
        job.save()

        return job.phase

    # check if the quota is exceeded
    if QueryJob.objects.get_size(job.owner) > get_quota(job.owner):
        job.phase = job.PHASE_ERROR
        job.error_summary = str(_('Quota is exceeded. Please remove some of your jobs.'))
        job.save()

        return job.phase

    # set database and start time
    job.start_time = now()

    job.pid = adapter.database.fetch_pid()
    job.actual_query = adapter.database.build_query(job.database_name, job.table_name, job.native_query, job.timeout, job.max_records)
    job.phase = job.PHASE_EXECUTING
    job.start_time = now()
    job.save()

    logger.info('run_query %s started' % job.id)

    # get the actual query and submit the job to the database
    try:
        # this is where the work ist done (and the time is spend)
        adapter.database.execute(job.actual_query)

    except (ProgrammingError, InternalError) as e:
        job.phase = job.PHASE_ERROR
        job.error_summary = str(e)
        logger.info('run_query %s failed (%s)' % (job.id, job.error_summary))

    except OperationalError as e:
        # load the job again and check if the job was killed
        job = QueryJob.objects.get(pk=job_id)

        if job.phase != job.PHASE_ABORTED:
            job.phase = job.PHASE_ERROR
            job.error_summary = str(e)
            logger.info('run_query %s failed (%s)' % (job.id, job.error_summary))

    else:
        # get additional information about the completed job
        job.phase = job.PHASE_COMPLETED
        logger.info('run_query %s completed' % job.id)

    finally:
        # get timing and save the job object
        job.end_time = now()
        job.execution_duration = (job.end_time - job.start_time).seconds

        # get additional information about the completed job
        if job.phase == job.PHASE_COMPLETED:
            job.nrows, job.size = adapter.database.fetch_stats(job.database_name, job.table_name)

            # fetch the metadata for the columns
            job.metadata['columns'] = adapter.database.fetch_columns(job.database_name, job.table_name)

            # fetch additional metadata from the metadata store
            for column in job.metadata['columns']:
                if column['name'] in job.metadata['display_columns']:

                    try:
                        database_name, table_name, column_name = job.metadata['display_columns'][column['name']]
                    except ValueError:
                        continue

                    try:
                        original_column = Column.objects.get(
                            name=column_name,
                            table__name=table_name,
                            table__database__name=database_name
                        )
                    except Column.DoesNotExist:
                        continue

                    column.update({
                        'description': original_column.description,
                        'unit': original_column.unit,
                        'ucd': original_column.ucd,
                        'utype': original_column.utype,
                        'principal': original_column.principal,
                        'indexed': False,
                        'std': original_column.std
                    })

        job.save()

    return job.phase


@shared_task(track_started=True, base=Task)
def create_download_file(file_name, format_key, database_name, table_name, metadata, status, empty):
    # always import daiquiri packages inside the task
    from daiquiri.core.adapter import get_adapter

    # get logger
    logger = logging.getLogger(__name__)

    # log start
    logger.info('create_download_file %s started' % file_name)

    # write file using the generator in the adapter
    with open(file_name, 'w') as f:
        for line in get_adapter().download.generate(format_key, database_name, table_name, metadata, status, empty):
            f.write(line)

    # log completion
    logger.info('create_download_file %s completed' % file_name)



from django.contrib import admin

from .models import Directory


class DirectoryAdmin(admin.ModelAdmin):
    search_fields = ('path', )
    list_display = ('path', )
    list_display_links = ('path', )


admin.site.register(Directory, DirectoryAdmin)



from rest_framework import viewsets, filters, status
from rest_framework.response import Response
from rest_framework.decorators import list_route
from rest_framework.permissions import IsAuthenticated
from rest_framework.authentication import SessionAuthentication, TokenAuthentication

from daiquiri.core.adapter import get_adapter
from daiquiri.core.viewsets import ChoicesViewSet
from daiquiri.core.permissions import HasModelPermission
from daiquiri.core.constants import LICENSE_CHOICES, ACCESS_LEVEL_CHOICES

from .models import Database, Table, Column, Function
from .serializers import (
    DatabaseSerializer,
    TableSerializer,
    ColumnSerializer,
    FunctionSerializer
)
from .serializers.management import (
    DatabaseSerializer as ManagementDatabaseSerializer,
    FunctionSerializer as ManagementFunctionSerializer
)
from .serializers.user import (
    DatabaseSerializer as UserDatabaseSerializer,
    FunctionSerializer as UserFunctionSerializer
)


class DatabaseViewSet(viewsets.ModelViewSet):
    permission_classes = (HasModelPermission, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = Database.objects.all()
    serializer_class = DatabaseSerializer

    filter_backends = (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter)
    filter_fields = ('name', 'access_level', 'metadata_access_level')
    search_fields = ('name', 'description')
    ordering_fields = ('name', 'access_level', 'metadata_access_level')

    def create(self, request, *args, **kwargs):

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        database = serializer.save()

        if request.data.get('discover'):
            adapter = get_adapter()

            for table_metadata in adapter.database.fetch_tables(database.name):
                table_metadata['database'] = database.id
                table_metadata['groups'] = [group.id for group in database.groups.all()]
                for key in ['license', 'access_level', 'metadata_access_level']:
                    table_metadata[key] = getattr(database, key)

                table_serializer = TableSerializer(data=table_metadata)
                if table_serializer.is_valid():
                    table = table_serializer.save()

                    for column_metadata in adapter.database.fetch_columns(database.name, table.name):
                        column_metadata['table'] = table.id
                        column_metadata['groups'] = [group.id for group in table.groups.all()]
                        for key in ['access_level', 'metadata_access_level']:
                            column_metadata[key] = getattr(table, key)

                        column_serializer = ColumnSerializer(data=column_metadata)
                        if column_serializer.is_valid():
                            column_serializer.save()

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    @list_route()
    def management(self, request):
        queryset = Database.objects.all()
        serializer = ManagementDatabaseSerializer(queryset, many=True)
        return Response(serializer.data)

    @list_route(methods=['get'], permission_classes=[])
    def user(self, request):
        # filter the databases which are published for the groups of the user
        queryset = Database.objects.filter_by_access_level(self.request.user)
        serializer = UserDatabaseSerializer(queryset, context={'request': request}, many=True)
        return Response(serializer.data)


class TableViewSet(viewsets.ModelViewSet):
    permission_classes = (HasModelPermission, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = Table.objects.all()
    serializer_class = TableSerializer

    filter_backends = (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter)
    filter_fields = ('name', 'access_level', 'metadata_access_level')
    search_fields = ('name', 'description')
    ordering_fields = ('name', 'access_level', 'metadata_access_level')

    def create(self, request, *args, **kwargs):

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        table = serializer.save()

        if request.data.get('discover'):
            adapter = get_adapter()

            for column_metadata in adapter.database.fetch_columns(table.database.name, table.name):
                column_metadata['table'] = table.id
                column_metadata['groups'] = [group.id for group in table.groups.all()]
                for key in ['access_level', 'metadata_access_level']:
                    column_metadata[key] = getattr(table, key)

                column_serializer = ColumnSerializer(data=column_metadata)
                if column_serializer.is_valid():
                    column_serializer.save()

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    @list_route(methods=['get'])
    def discover(self, request):
        database_name = request.GET.get('database')
        table_name = request.GET.get('table')

        if database_name and table_name:
            return Response([get_adapter().database.fetch_table(database_name, table_name)])
        else:
            return Response([])


class ColumnViewSet(viewsets.ModelViewSet):
    permission_classes = (HasModelPermission, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = Column.objects.all()
    serializer_class = ColumnSerializer

    filter_backends = (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter)
    filter_fields = ('name', 'access_level', 'metadata_access_level')
    search_fields = ('name', 'description')
    ordering_fields = ('name', 'access_level', 'metadata_access_level')

    @list_route(methods=['get'])
    def discover(self, request):
        database_name = request.GET.get('database')
        table_name = request.GET.get('table')
        column_name = request.GET.get('column')

        if database_name and table_name:
            return Response([get_adapter().database.fetch_column(database_name, table_name, column_name)])
        else:
            return Response([])


class FunctionViewSet(viewsets.ModelViewSet):
    permission_classes = (HasModelPermission, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = Function.objects.all()
    serializer_class = FunctionSerializer

    filter_backends = (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter)
    filter_fields = ('name', 'access_level', 'metadata_access_level')
    search_fields = ('name', 'description')
    ordering_fields = ('name', 'access_level', 'metadata_access_level')

    @list_route(methods=['get'])
    def management(self, request):
        queryset = Function.objects.all()
        serializer = ManagementFunctionSerializer(queryset, many=True)
        return Response(serializer.data)

    @list_route(methods=['get'], permission_classes=[])
    def user(self, request):
        queryset = Function.objects.filter_by_access_level(self.request.user)
        serializer = UserFunctionSerializer(queryset, many=True)
        return Response(serializer.data)


class TableTypeViewSet(ChoicesViewSet):
    permission_classes = (IsAuthenticated, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = Table.TYPE_CHOICES


class LicenseViewSet(ChoicesViewSet):
    permission_classes = (IsAuthenticated, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = LICENSE_CHOICES


class AccessLevelViewSet(ChoicesViewSet):
    permission_classes = (IsAuthenticated, )
    authentication_classes = (SessionAuthentication, TokenAuthentication)

    queryset = ACCESS_LEVEL_CHOICES



import os

from django.conf import settings

from daiquiri.core.adapter import get_adapter
from daiquiri.metadata.models import Database, Table, Column
from daiquiri.files.models import Directory
from daiquiri.files.utils import normalize_file_path
from daiquiri.query.models import QueryJob
from daiquiri.query.utils import get_user_database_name


def get_columns(user, database_name, table_name):

    user_database_name = get_user_database_name(user)

    if database_name == user_database_name:
        # get the job fetch the columns
        try:
            job = QueryJob.objects.filter_by_owner(user).exclude(phase=QueryJob.PHASE_ARCHIVED).get(
                database_name=database_name,
                table_name=table_name
            )
            return job.metadata['columns']
        except QueryJob.DoesNotExist:
            return []

    else:
        # check permissions on the database
        try:
            database = Database.objects.filter_by_access_level(user).get(name=database_name)
        except Database.DoesNotExist:
            return []

        # check permissions on the table
        try:
            table = Table.objects.filter_by_access_level(user).filter(database=database).get(name=table_name)
        except Table.DoesNotExist:
            return []

        # get columns for this table
        return Column.objects.filter_by_access_level(user).filter(table=table).values()


def get_column(user, database_name, table_name, column_name):

    columns = get_columns(user, database_name, table_name)

    try:
        return [column for column in columns if column['name'] == column_name][0]
    except IndexError:
        # column_name is not in columns
        return None


def get_archive_files(user, database_name, table_name, column_name):
    files = []

    if database_name and table_name and column_name:
        # get directories this user has access to
        directories = Directory.objects.filter_by_access_level(user)

        # get columns of this table the user is allowed to access
        column = get_column(user, database_name, table_name, column_name)
        if column:
            # get the filenames
            adapter = get_adapter()
            count = adapter.database.count_rows(database_name, table_name)
            rows = adapter.database.fetch_rows(database_name, table_name, [column['name']], page_size=count)

            for row in rows:
                for file_path in row:
                    for directory in directories:
                        file_path = normalize_file_path(directory.path, file_path)
                        if os.path.isfile(os.path.join(directory.path, file_path)):
                            files.append((directory.path, file_path))

    return files


def get_archive_file_name(user, table_name, column_name):
    if not user or user.is_anonymous():
        username = 'anonymous'
    else:
        username = user.username

    directory_name = os.path.join(settings.SERVE_DOWNLOAD_DIR, username)
    return os.path.join(directory_name, '%s_%s.zip' % (table_name, column_name))



import os
import mock

from django.conf import settings
from django.test import TestCase

from test_generator.viewsets import (
    TestModelViewsetMixin,
    TestListViewsetMixin,
    TestViewsetMixin
)

from daiquiri.core.adapter import get_adapter

from ..models import QueryJob, Example

adapter = get_adapter()

mock_execute = mock.Mock()

mock_fetch_stats = mock.Mock()
mock_fetch_stats.return_value = [0, 0]


class QueryViewsetTestCase(TestCase):

    fixtures = (
        'auth.json',
        'metadata.json',
        'jobs.json',
        'queryjobs.json'
    )

    users = (
        ('admin', 'admin'),
        ('user', 'user'),
        ('evil', 'evil'),
        ('anonymous', None),
    )

    status_map = {
        'list_viewset': {
            'admin': 200, 'user': 200, 'evil': 200, 'anonymous': 403
        }
    }


class StatusTests(TestListViewsetMixin, QueryViewsetTestCase):

    url_names = {
        'viewset': 'query:status'
    }


class FormTests(TestListViewsetMixin, QueryViewsetTestCase):

    url_names = {
        'viewset': 'query:form'
    }


class DropdownTests(TestListViewsetMixin, QueryViewsetTestCase):

    url_names = {
        'viewset': 'query:dropdown'
    }


class JobTests(TestViewsetMixin, QueryViewsetTestCase):

    instances = QueryJob.objects.filter(owner__username='user')

    url_names = {
        'viewset': 'query:job'
    }

    status_map = {
        'list_viewset': {
            'admin': 200, 'user': 200, 'evil': 200, 'anonymous': 403
        },
        'detail_viewset': {
            'admin': 404, 'user': 200, 'evil': 404, 'anonymous': 403
        },
        'create_viewset': {
            'admin': 201, 'user': 201, 'evil': 201, 'anonymous': 403
        },
        'update_viewset': {
            'admin': 404, 'user': 200, 'evil': 404, 'anonymous': 403
        },
        'delete_viewset': {
            'admin': 404, 'user': 204, 'evil': 404, 'anonymous': 403
        },
        'abort_viewset': {
            'admin': 404, 'user': 200, 'evil': 404, 'anonymous': 403
        },
        'download_completed_viewset': {
            'admin': 404, 'user': 200, 'evil': 404, 'anonymous': 403
        },
        'download_executing_viewset': {
            'admin': 404, 'user': 400, 'evil': 404, 'anonymous': 403
        },
        'stream_completed_viewset': {
            'admin': 404, 'user': 200, 'evil': 404, 'anonymous': 403
        },
        'stream_executing_viewset': {
            'admin': 404, 'user': 400, 'evil': 404, 'anonymous': 403
        },
    }

    def _test_list_viewset(self, username):
        self.assert_list_viewset(username)

    def _test_detail_viewset(self, username):
        for instance in self.instances:
            self.assert_detail_viewset(username, kwargs={'pk': instance.pk})

    @mock.patch.object(adapter.database, 'execute', mock_execute)
    @mock.patch.object(adapter.database, 'fetch_stats', mock_fetch_stats)
    def _test_create_viewset(self, username):
        for example in Example.objects.all():
            self.assert_create_viewset(username, data={
                'query_language': example.query_language,
                'query': example.query
            })

    @mock.patch.object(adapter.database, 'execute', mock_execute)
    def _test_update_viewset(self, username):
        for instance in self.instances:
            self.assert_update_viewset(username, kwargs={
                'pk': instance.pk
            }, data={
                'table_name': instance.table_name + '_renamed'
            })

    @mock.patch.object(adapter.database, 'execute', mock_execute)
    def _test_delete_viewset(self, username):
        for instance in self.instances:
            self.assert_delete_viewset(username, kwargs={
                'pk': instance.pk
            })

    def _test_abort_viewset(self, username):
        for instance in self.instances:
            self.assert_viewset('abort_viewset', 'put', 'abort', username, kwargs={
                'pk': instance.pk
            })

    def _test_download_completed_viewset(self, username):

        instance = self.instances.filter(phase='COMPLETED').first()
        self.assert_download_viewset('download_completed_viewset', instance, username)

    def _test_download_executing_viewset(self, username):

        instance = self.instances.filter(phase='EXECUTING').first()
        self.assert_download_viewset('download_executing_viewset', instance, username)

    def assert_download_viewset(self, key, instance, username):

        for format_key in ('csv', 'votable', 'votable-binary', 'votable-binary2'):
            file_name = self.get_download_file_name(instance, format_key)

            for method in ['get', 'put']:
                try:
                    os.remove(os.path.join(settings.QUERY_DOWNLOAD_DIR, file_name))
                except OSError:
                    pass

                # file is not existing yet
                self.assert_viewset(key, method, 'download', username, kwargs={
                    'pk': instance.pk,
                    'format_key': format_key
                })

                # file exists
                self.assert_viewset(key, method, 'download', username, kwargs={
                    'pk': instance.pk,
                    'format_key': format_key
                })

    def _test_stream_completed_viewset(self, username):

        instance = self.instances.filter(phase='COMPLETED').first()
        self.assert_stream_viewset('stream_completed_viewset', instance, username)

    def _test_stream_executing_viewset(self, username):

        instance = self.instances.filter(phase='EXECUTING').first()
        self.assert_stream_viewset('stream_executing_viewset', instance, username)

    def assert_stream_viewset(self, key, instance, username):

        for format_key in ('csv', 'votable', 'votable-binary', 'votable-binary2'):
            file_name = self.get_download_file_name(instance, format_key)

            try:
                os.remove(os.path.join(settings.QUERY_DOWNLOAD_DIR, file_name))
            except OSError:
                pass

            # file is not existing yet
            self.assert_viewset(key, 'get', 'stream', username, kwargs={
                'pk': instance.pk,
                'format_key': format_key
            })

            # file exists
            self.assert_viewset(key, 'get', 'stream', username, kwargs={
                'pk': instance.pk,
                'format_key': format_key
            })

    def get_download_file_name(self, instance, format_key):
        file_name =  '%(username)s/%(table_name)s' % {
            'username': instance.owner.username,
            'table_name': instance.table_name
        }

        if format_key == 'csv':
            return file_name + '.csv'
        elif format_key == 'votable':
            return file_name + '.votable.xml'
        elif format_key == 'votable-binary':
            return file_name + '.votable.binary.xml'
        elif format_key == 'votable-binary2':
            return file_name + '.votable.binary2.xml'

class ExampleViewsetTestCase(TestCase):

    fixtures = (
        'auth.json',
        'metadata.json'
    )

    users = (
        ('admin', 'admin'),
        ('user', 'user'),
        ('anonymous', None),
    )

    status_map = {
        'list_viewset': {
            'admin': 200, 'user': 403, 'anonymous': 403
        },
        'detail_viewset': {
            'admin': 200, 'user': 403, 'anonymous': 403
        },
        'create_viewset': {
            'admin': 201, 'user': 403, 'anonymous': 403
        },
        'update_viewset': {
            'admin': 200, 'user': 403, 'anonymous': 403
        },
        'delete_viewset': {
            'admin': 204, 'user': 403, 'anonymous': 403
        },
        'user_viewset': {
            'admin': 200, 'user': 200, 'anonymous': 403
        }
    }


class ExampleTests(TestModelViewsetMixin, ExampleViewsetTestCase):

    instances = Example.objects.all()
    url_names = {
        'viewset': 'query:example'
    }

    def _test_user_viewset(self, username):
        self.assert_viewset('user_viewset', 'get', 'user', username)
